/*
    © Copyright 2016 Wouter Meulemans 
    
    This file is part of SMWGComputation.

    SMWGComputation is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SMWGComputation is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with SMWGComputation.  If not, see <http://www.gnu.org/licenses/>.
*/

/* global data, batch, visual, debug */
/* global batchLikeOutput */
/* global fbs */

function analyzeMap(suppressclear) {
    if (!hasValue(data.regions) && (!batch.active || !batch.writeheaders)) {
        return;
    }

    if (!batch.active && !suppressclear) {
        debug.clear();
    }

    if (batch.active && !batch.writeheaders) {
        var tb = textBuilder();
        appendMapGeometry(tb);
        fbs.hiddendoc(tb.string, batch.map + ".txt", "text/plain", "maps");
    }

    var frame = getEmptyBox();
    frame.left = frame.top = 0;
    frame.right = frame.bottom = 500;

    var box;
    if (hasValue(data.regions)) {
        box = getBoundingBoxOfRegions(data.regions, frame);
        data.region_aspectRatio = box.width() / box.height();
    } else {
        box = null;
        data.region_aspectRatio = 1;
    }

    var mapmeasures = [
        {
            name: "Regions",
            summary: false,
            unit: "",
            measure: function () {
                return data.regions.length;
            }
        },
        {
            name: "Whitespace",
            summary: false,
            unit: "%",
            measure: function () {
                var totarea = 0;
                data.regions.forEach(function (R) {
                    totarea += R.area;
                });
                return 100 - 100 * totarea / (box.width() * box.height());
            }
        },
        {
            name: "Aspect ratio",
            summary: false,
            unit: "",
            measure: function () {
                return box.width() / box.height();
            }
        },
        {
            name: "Area",
            summary: true,
            unit: "px^2",
            measure: function () {
                var areas = [];
                data.regions.forEach(function (R) {
                    areas.push(R.area);
                });
                return summary(areas);
            }
        },
        {
            name: "Neighbors",
            summary: true,
            unit: "",
            measure: function () {
                var neighbors = [];
                data.regions.forEach(function (R) {
                    neighbors.push(R.topology.length);
                });
                return summary(neighbors);
            }
        },
        {
            name: "Squareness",
            summary: true,
            unit: "",
            measure: function () {
                var squareness = [];
                data.regions.forEach(function (R) {
                    var Rbox = getBoundingBoxOfParts(R.parts);
                    var d = Rbox.height() - Rbox.width();
                    if (d > 0) {
                        Rbox.left -= d / 2;
                        Rbox.right += d / 2;
                    } else {
                        Rbox.top += d / 2;
                        Rbox.bottom -= d / 2;
                    }
                    var Rsqr = 1 - R.area / (Rbox.width() * Rbox.height());
                    squareness.push(Rsqr);
                });
                return summary(squareness);
            }
        }
    ];

    outputAnalysis(mapmeasures, 'map');
}

function analyzeGrid(suppressclear) {
    if (!hasValue(data.grid) && (!batch.active || !batch.writeheaders)) {
        return;
    }

    if (!batch.active && !suppressclear) {
        debug.clear();
    }

    var box = hasValue(data.grid) ? getBoundingBoxOfParts(data.gridcells) : null;

    var gridmeasures = [
        {
            name: "Rows",
            summary: false,
            unit: "",
            measure: function () {
                return data.grid[0].length;
            }
        },
        {
            name: "Columns",
            summary: false,
            unit: "",
            measure: function () {
                return data.grid.length;
            }
        },
        {
            name: "Cells",
            summary: false,
            unit: "",
            measure: function () {
                return data.gridcells.length;
            }
        },
        {
            name: "Whitespace",
            summary: false,
            unit: "%",
            measure: function () {
                return 100 - 100 * data.regions.length / data.gridcells.length;
            }
        },
        {
            name: "Aspect ratio (grid)",
            summary: false,
            unit: "",
            measure: function () {
                return box.width() / box.height();
            }
        },
        {
            name: "Aspect ratio (cell)",
            summary: false,
            unit: "",
            measure: function () {
                var cellbox = getBoundingBoxOfGeometry(data.gridcells[0].geometry);
                return cellbox.width() / cellbox.height();
            }
        },
        {
            name: "Area",
            summary: false,
            unit: "px^2",
            measure: function () {
                return box.height() * box.width();
            }
        }
    ];

    if (batch.active && !batch.writeheaders) {
        var tb = textBuilder();
        appendGridGeometry(tb);
        fbs.hiddendoc(tb.string, batch.map + "_" + batch.ws + ".txt", "text/plain", "grids");
    }

    outputAnalysis(gridmeasures, 'grid');
}

function analyzeMapAndGrid() {
    if (!hasValue(data.regions)) {
        return;
    }
    if (!hasValue(data.gridcells)) {
        return;
    }

    function point_in_polygon(p, polygon) {
        var totangle = 0;
        var prev = polygon[polygon.length - 1];
        for (var i = 0; i < polygon.length; i++) {
            var curr = polygon[i];
            totangle += computeSignedAngle(p, prev, p, curr);
            prev = curr;
        }
        // its either 0 or a multiple of 2 PI
        // 0  -> outside
        // !0 -> inside
        return !(-0.5 < totangle && totangle < 0.5);
    }

    function point_is_covered(p) {
        for (var ri = 0; ri < data.regions.length; ri++) {
            var region = data.regions[ri];
            if (region.box.contains(p)) {
                for (var pi = 0; pi < region.parts.length; pi++) {
                    var part = region.parts[pi];
                    if (part.box.contains(p) && point_in_polygon(p, part.geometry)) {
                        return true;
                    }
                }
            }
        }
        return false;
    }
    
    function compute_overlap(cell) {
        var nH = Math.ceil(750 / data.grid.length);
        var nV = Math.ceil(750 / data.grid[0].length);

        var contained = 0;
        var incell = 0;

        var box = getBoundingBoxOfGeometry(cell.geometry);

        var bh = box.left;
        var sh = box.width() / (nH + 1);
        var bv = box.top;
        var sv = box.height() / (nV + 1);

        for (var ih = 1; ih <= nH; ih++) {
            for (var iv = 1; iv <= nV; iv++) {
                var p = [
                    bh + sh * ih,
                    bv + sv * iv
                ];
                if (point_in_polygon(p, cell.geometry)) {
                    incell++;
                    if (point_is_covered(p)) {
                        contained++;
                    }
                }
            }
        }

        cell.overlap = contained / incell;
    }
    for (var ci = 0; ci < data.gridcells.length; ci++) {
        compute_overlap(data.gridcells[ci]);
    }

    var maxspread = 0;
    for (var ci = 0; ci < data.gridcells.length; ci++) {
        var cell = data.gridcells[ci];

        var nearest = cells_by_distance(cell);
        var sumweight = 0;
        cell.spread = 0;
        var ni = 0;
        var lastdist;
        while (sumweight < 0.9999999 && ni < nearest.length) {
            cell.spread += nearest[ni].cell.overlap * nearest[ni].distance;
            lastdist = nearest[ni].distance;
            sumweight += nearest[ni].cell.overlap;

            ni++;
        }
        cell.spread -= Math.max(0, sumweight - 1) * lastdist;

        maxspread = Math.max(maxspread, cell.spread);
    }

    if (!batch.active) {
        visual.updateJointAux(maxspread);
    }
}

function appendGridGeometry(builder) {
    data.gridcells.forEach(function (cell) {
        builder.appendLine(cell.column + " " + cell.row);
        var first = true;
        cell.geometry.forEach(function (p) {
            if (!first) {
                builder.append(" ");
            } else {
                first = false;
            }
            builder.append(p[0] + " " + p[1]);
        });
        builder.appendLine("");
    });
}

function appendMapGeometry(builder) {
    data.regions.forEach(function (region) {
        builder.appendLine(region.name);
        region.parts.forEach(function (part) {
            var first = true;
            part.geometry.forEach(function (p) {
                if (!first) {
                    builder.append(" ");
                } else {
                    first = false;
                }
                builder.append(p[0] + " " + p[1]);
            });
            builder.appendLine("");
        });
    });
}

function outputAnalysis(analysismeasures, type) {

    if (batch.active || batchLikeOutput) {
        var builder;
        if (!batch.active) {
            builder = debug;
        } else {
            switch (type) {
                case 'map':
                    builder = batch.map_data;
                    break;
                case 'grid':
                    builder = batch.grid_data;
                    break;
            }
        }

        if (!batch.active || batch.writeheaders) {
            switch (type) {
                case 'map':
                    builder.append("map");
                    break;
                case 'grid':
                    builder.append("map\twhitespace");
                    break;
            }

            analysismeasures.forEach(function (measure) {
                if (measure.summary) {
                    builder.append("\t" + measure.name + ":mean");
                    builder.append("\t" + measure.name + ":stddev");
                    builder.append("\t" + measure.name + ":min");
                    builder.append("\t" + measure.name + ":25pct");
                    builder.append("\t" + measure.name + ":median");
                    builder.append("\t" + measure.name + ":75pct");
                    builder.append("\t" + measure.name + ":max");
                } else {
                    builder.append("\t" + measure.name);
                }
            });

            builder.appendLine("");

        }

        if (!batch.active || !batch.writeheaders) {
            switch (type) {
                case 'map':
                    builder.append(batch.active ? batch.map : "MAP");
                    break;
                case 'grid':
                    builder.append(batch.active ? batch.map + "\t" + batch.ws : "MAP\tWS");
                    break;
            }

            analysismeasures.forEach(function (measure) {
                var value = measure.measure();
                if (measure.summary) {
                    builder.append("\t" + value.avg);
                    builder.append("\t" + value.stddev);
                    builder.append("\t" + value.min);
                    builder.append("\t" + value.pct25);
                    builder.append("\t" + value.median);
                    builder.append("\t" + value.pct75);
                    builder.append("\t" + value.max);
                } else {
                    builder.append("\t" + value);
                }
            });

            builder.appendLine("");
        }
    } else {

        analysismeasures.forEach(function (measure) {
            var value = measure.measure();
            if (measure.summary) {
                debug.appendLine(measure.name + " - mean: \t" + value.avg + "\t " + measure.unit);
                debug.appendLine(measure.name + " - stddev: \t" + value.stddev + "\t " + measure.unit);
                debug.appendLine(measure.name + " - min: \t" + value.min + "\t " + measure.unit);
                debug.appendLine(measure.name + " - 25pct: \t" + value.pct25 + "\t " + measure.unit);
                debug.appendLine(measure.name + " - median: \t" + value.median + "\t " + measure.unit);
                debug.appendLine(measure.name + " - 75pct: \t" + value.pct75 + "\t " + measure.unit);
                debug.appendLine(measure.name + " - max: \t" + value.max + "\t " + measure.unit);
            } else {
                debug.appendLine(measure.name + ": \t" + value + "\t " + measure.unit);
            }
        });

    }
}
