/*
    © Copyright 2016 Wouter Meulemans 
    
    This file is part of SMWGComputation.

    SMWGComputation is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SMWGComputation is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with SMWGComputation.  If not, see <http://www.gnu.org/licenses/>.
*/

/* global d3 */
/* global batch, debug, visual */
/* global rectMath, hexHorzMath, hexVertMath */

var data = (function () {
    var dt = {
        // public variables
        spatialdata: true,
        regions: null,
        region_aspectRatio: null,
        gridcells: null,
        grid: null,
        gridmath: null,
        // public methods
        renewMap: _reload,
        renewGrid: _renewgrid,
        clearAssignment: _clear,
        setFromWhitespace: _setFromWhitespace,
        getGridCell: _getGridCell,
        isOnGrid: _isOnGrid,
        loadassignment: _loadassignment
    };

    function _reload(givenmap, callback, prefix) {
        if (batch.active && batch.writeheaders) {
            batch.settings_data.appendLine("Derived topology\ttrue");
            return;
        }

        if (!hasValue(prefix)) {
            prefix = "";
        }

        var map = !hasValue(givenmap) ? d3.select("#map-file").property("value") : givenmap;

        _clear();

        if (map === 'iris-petal') {
            _readIrisData('Petal length', 'Petal width', 'Sepal width', 'Sepal length', callback);
        } else if (map === 'iris-sepal') {
            _readIrisData('Sepal length', 'Sepal width', 'Petal width', 'Petal length', callback);
        } else {
            _readMapData(map, prefix, callback);
        }
    }

    function _clear() {
        if (!batch.active)
            visual.clearmouseover();

        // clear current assignment
        if (hasValue(dt.regions))
            dt.regions.forEach(function (region) {
                region.cell = undefined;
            });
        if (hasValue(dt.gridcells))
            dt.gridcells.forEach(function (cell) {
                cell.region = undefined;
            });

        if (!batch.active) {
            visual.clearGrid();
        }
    }

    function _readIrisData(xfield, yfield, xellipse, yellipse, callback) {

        dt.regions = [];
        dt.spatialdata = false;
        d3.tsv("data/iris.tsv", function (error, tsvdata) {

            var minX = Number.POSITIVE_INFINITY;
            var maxX = Number.NEGATIVE_INFINITY;
            var minY = Number.POSITIVE_INFINITY;
            var maxY = Number.NEGATIVE_INFINITY;
            tsvdata.forEach(function (measurement) {
                minX = Math.min(minX, Number(measurement[xfield]));
                maxX = Math.max(maxX, Number(measurement[xfield]));
                minY = Math.min(minY, Number(measurement[yfield]));
                maxY = Math.max(maxY, Number(measurement[yfield]));
            });

            function toCoords(m) {
                var xfrac = (Number(m[xfield]) - minX) / (maxX - minX);
                var yfrac = (Number(m[yfield]) - minY) / (maxY - minY);
                return [
                    30 + 440 * xfrac,
                    30 + 440 * yfrac
                ];
            }

            var deltaX = 440 * 0.1 / (maxX - minX);
            var deltaY = 440 * 0.1 / (maxY - minY);

            var r = 4;

            tsvdata.forEach(function (measurement, index) {
                var region = {};
                region.name = "M" + measurement['Measurement'];
                region.index = index;
                region.area = 4 * r * r;
                region.center = toCoords(measurement);
                region.ellipse = [Number(measurement[xellipse]), Number(measurement[yellipse])];
                region.color = measurement.Species === 'I. setosa'
                        ? "rgb(228,26,28)"
                        : (measurement.Species === 'I. versicolor'
                                ? "rgb(55,126,184)"
                                : "rgb(77,175,74)");
                region.topology = [];
                region.parts = [];

                var part = {};
                region.parts.push(part);
                part.topology = [];
                part.geometry = [
                    [region.center[0] - r, region.center[1] - r],
                    [region.center[0] + r, region.center[1] - r],
                    [region.center[0] + r, region.center[1] + r],
                    [region.center[0] - r, region.center[1] + r]
                ];
                part.center = region.center;
                part.area = region.area;
                part.box = getBoundingBoxOfGeometry(part.geometry);

                region.box = getBoundingBoxOfParts(region.parts);

                dt.regions.push(region);
            });

			// NB: no topology!
			
            analyzeMap();
            analyzeMapAndGrid();
            if (!batch.active)
                visual.visualizeMap();
            if (callback) {
                callback();
            }
        });
    }

    function _constructPolygon(part, decomposedmap) {

        var result = [];

        for (var tpindex = 0; tpindex < part.length; tpindex++) {
            var topochain = part[tpindex];
            if (topochain.polyline === undefined) {
                // polygon!
                var polygon = decomposedmap.polygons[topochain.polygon];
                if (topochain.reversed) {
                    for (var i = polygon.length - 1; i >= 0; i--) {
                        result.push(polygon[i]);
                    }
                } else {
                    for (var i = 0; i < polygon.length; i++) {
                        result.push(polygon[i]);
                    }
                }
            } else {
                // NB: skip over first intentional to make a simple polygon
                var chain = decomposedmap.polylines[topochain.polyline];
                if (topochain.reversed) {
                    for (var i = chain.length - 2; i >= 0; i--) {
                        result.push(chain[i]);
                    }
                } else {
                    for (var i = 1; i < chain.length; i++) {
                        result.push(chain[i]);
                    }
                }
            }
        }

        return result;
    }

    function _readMapData(map, prefix, callback) {
        dt.spatialdata = true;

        function on_done() {
            analyzeMap();
            analyzeMapAndGrid();
            if (!batch.active)
                visual.visualizeMap();
            if (callback)
                callback();
        }

        dt.regions = [];
        d3.json("data/" + prefix + map + "_decomposed.json", function (decomposedmap) {

            decomposedmap.regions.forEach(function (inputregion, index) {

                var region = {};
                region.name = inputregion.name;
                region.index = index;

                region.area = 0;
                region.center = [0, 0];

                region.topology = [];

                region.parts = [];
                inputregion.parts.forEach(function (inputpart) {
                    var part = {};

                    part.topology = inputpart;
                    part.geometry = _constructPolygon(inputpart, decomposedmap);

                    part.center = computeMidpoint(part.geometry);
                    part.area = computeArea(part.geometry);
                    part.box = getBoundingBoxOfGeometry(part.geometry);

                    if (isNaN(part.center[0])) {
                        // part of negible area
                        return;
                    }

                    region.parts.push(part);

                    region.area += part.area;
                    region.center[0] += part.area * part.center[0];
                    region.center[1] += part.area * part.center[1];
                });

                if (region.parts.length === 0) {
                    console.log(region.name + "has no significantly sized part");
                }

                region.center[0] /= region.area;
                region.center[1] /= region.area;

                region.box = getBoundingBoxOfParts(region.parts);

                dt.regions.push(region);
            });

			dt.regions.forEach(function (region) {
				region.parts.forEach(function (part) {
					part.topology.forEach(function (tp) {
						if (tp.neighbor !== null) {
							var nbr = dt.regions[tp.neighbor[0]];

							var len = 0;
							var chain = decomposedmap.polylines[tp.polyline];
							for (var i = 1; i < chain.length; i++) {
								len += computeEuclideanDistance(chain[i - 1], chain[i]);
							}

							// check if it doesnt exist yet
							var novel = true;
							region.topology.forEach(function (rtp) {
								if (rtp.neighbor === nbr) {
									rtp.strength += len;
									novel = false;
								}
							});
							if (novel) {
								region.topology.push({neighbor: nbr, category: "T", strength: len});
							}
						}
					});
				});
			});

			var maxstrength = 0;
			dt.regions.forEach(function (region) {
				region.topology.forEach(function (tp) {
					maxstrength = Math.max(maxstrength, tp.strength);
				});
			});

			dt.regions.forEach(function (region) {
				region.topology.forEach(function (tp) {
					tp.strength /= maxstrength;
				});
			});

			on_done();            
        });
    }

    function _renewgrid(cols, rows) {

        var shape = d3.select("#grid-shape").property("value");
        var gridsize = d3.select("#grid-size").property("value");
        var cellratio = d3.select("#grid-cellratio").property("value");
        var ratio = Number(d3.select("#grid-ratio").property("value"));

        if (batch.active && batch.writeheaders) {
            batch.settings_data.appendLine("Grid shape\t" + shape);
            batch.settings_data.appendLine("Grid size\t" + gridsize);
            batch.settings_data.appendLine("Grid aspect ratio\t" + ratio);
            batch.settings_data.appendLine("Cell aspect ratio\t" + cellratio);
            return;
        }

        var ncolumns = cols === undefined ? Number(d3.select("#grid-columns").property("value")) : cols;
        var nrows = rows === undefined ? Number(d3.select("#grid-rows").property("value")) : rows;

        switch (shape) {
            default:
            case "square":
                dt.gridmath = rectMath;
                break;
            case "hexh":
                dt.gridmath = hexHorzMath;
                break;
            case "hexv":
                dt.gridmath = hexVertMath;
                break;
        }

        var fixratio = cellratio === 'fixed';

        var gridbase;
        switch (gridsize) {
            default:
            case "spanning":
                var totwidth, totheight;
                if (ratio > 1) {
                    totwidth = 480;
                    totheight = 480 / ratio;
                } else {
                    totwidth = 480 * ratio;
                    totheight = 480;
                }
                gridbase = dt.gridmath.computeSpanningBase(totwidth, totheight, ncolumns, nrows, fixratio);
                break;
            case "same-area":
                var cellarea = 0;
                dt.regions.forEach(function (region) {
                    cellarea += region.area;
                });
                cellarea /= dt.regions.length;
                gridbase = dt.gridmath.computeAreaBase(ncolumns, nrows, cellarea, fixratio ? 1 : ratio);
                break;
        }

        dt.gridcells = [];
        dt.grid = [];
        for (var col = 0; col < ncolumns; col++) {
            dt.grid[col] = [];
            for (var row = 0; row < nrows; row++) {
                var cell = {};
                cell.index = dt.gridcells.length;
                cell.column = col;
                cell.row = row;
                cell.geometry = dt.gridmath.getCellGeometry(col, row, gridbase);
                cell.center = computeMidpoint(cell.geometry);
                cell.area = computeArea(cell.geometry);

                var box = getBoundingBoxOfGeometry(cell.geometry);
                cell.width = box.width();
                cell.height = box.height();

                dt.grid[col][row] = cell;
                dt.gridcells.push(cell);
                // NB: procedure will make gridcells contain row after row
            }
        }

        _clear();
        analyzeMapAndGrid();
        if (!batch.active)
            visual.visualizeGrid();
    }


    function _setFromWhitespace(whitespacefraction) {

        var wsmethod = d3.select("#grid-whitespace-method").property("value");
        var slack = Number(d3.select("#grid-whitespace-slack").property("value")) / 100;

        if (batch.active && batch.writeheaders) {
            batch.settings_data.append("Whitespace method\t" + wsmethod);
            if (wsmethod === 'min-ar') {
                batch.settings_data.append(":" + slack);
            }
            batch.settings_data.appendLine("");
            return;
        }

        var N = dt.regions.length;
        var A = dt.region_aspectRatio;
        var F = whitespacefraction === undefined ? d3.select("#grid-whitespace").property("value") : whitespacefraction;

        var T = N / (1 - F);

        var R;
        var C;

        switch (wsmethod) {
            default:
            case "square":
                R = C = Math.ceil(Math.sqrt(T));
                break;
            case "min-ar":
                if (A < 1) {
                    A = A * slack;
                } else {
                    A = A / slack;
                }

                if (A <= 1) {
                    // C/R <= 1, 
                    // A <= C/R, 
                    // so AR <= C <= R
                    // find integer R,C 
                    // minimizing R*C - T
                    // s.t. R*C >= T
                    //      AR <= C <= R
                    // min value for R is when C = R, implying R = ceil(sqrt(T))
                    // max value for R is when C = AR, implying R = ceil(sqrt(T/A))                        
                    var D = Number.POSITIVE_INFINITY;
                    for (var r = Math.ceil(Math.sqrt(T));
                            r <= Math.ceil(Math.sqrt(T / A)); r++) {
                        // given r, what is smallest value for c?
                        var c = Math.max(Math.ceil(T / r), Math.ceil(A * r));
                        var d = r * c - T;
                        if (d < D) {
                            D = d;
                            R = r;
                            C = c;
                        }
                    }
                } else {
                    // 1 <= C/R, 
                    // C/R <= A, 
                    // so R <= C <= AR
                    // find integer R,C 
                    // minimizing R*C - T
                    // s.t. R*C >= T
                    //      C/A <= R <= C
                    // min value for C is when R = C, implying C = ceil(sqrt(T))
                    // max value for C is when R = C/A, implying C = ceil(sqrt(T*A))                        
                    var D = Number.POSITIVE_INFINITY;
                    for (var c = Math.ceil(Math.sqrt(T));
                            c <= Math.ceil(Math.sqrt(T * A)); c++) {
                        // given c, what is smallest value for r?
                        var r = Math.max(Math.ceil(T / c), Math.ceil(c / A));
                        var d = r * c - T;
                        if (d < D) {
                            D = d;
                            R = r;
                            C = c;
                        }
                    }
                }
                break;
            case "any-ar":
                // C/R <= 1
                // so C <= R
                // find integer R,C 
                // minimizing R*C - T
                // s.t. R*C >= T
                //      1 <= C <= R
                // min value for R is when C = R, implying R = ceil(sqrt(T))
                // max value for R is when C = 1, implying R = T                        
                var D = Number.POSITIVE_INFINITY;
                for (var r = Math.ceil(Math.sqrt(T));
                        r <= T; r++) {
                    // given r, what is smallest value for c?
                    var c = Math.ceil(T / r);
                    var d = r * c - T;
                    if (d < D) {
                        D = d;
                        R = r;
                        C = c;
                    }
                }

                if (A > 1) {
                    var t = R;
                    R = C;
                    C = t;
                }
                break;
        }

        if (!batch.active) {
            d3.select("#grid-columns").property("value", C);
            d3.select("#grid-rows").property("value", R);
        }

        _renewgrid(C, R);

        return 1 - N / (R * C);
    }

    function _isOnGrid(column, row) {
        return 0 <= column && column < dt.grid.length
                && 0 <= row && row < dt.grid[0].length;
    }

    function _getGridCell(column, row) {
        if (_isOnGrid(column, row)) {
            return dt.grid[column][row];
        } else {
            return null;
        }
    }


    function _loadassignment(given, callback, prefix, keepgrid, offX, offY) {

        batchLikeOutput = true;

        if (!hasValue(prefix)) {
            prefix = "";
        }

        if (!hasValue(offX)) {
            offX = 0;
        }
        if (!hasValue(offY)) {
            offY = 0;
        }

        var assignment;
        if (!hasValue(given)) {
            assignment = d3.select("#assign-file").property("value");
            if (assignment.length < 1) {
                var map = d3.select("#map-file").property("value");
                assignment = map + "_gridcells";
            }
        } else {
            assignment = given;
        }

        debug.clear();

        d3.tsv("data/" + prefix + assignment + ".tsv", function (error, data) {
            if (!keepgrid) {
                var maxX = 0;
                var maxY = 0;
                data.forEach(function (assignment) {
                    maxX = Math.max(maxX, assignment.gridX);
                    maxY = Math.max(maxY, assignment.gridY);
                });

                renewgrid(maxX + 1, maxY + 1);
            }

            data.gridcells.forEach(function (cell) {
                cell.region = undefined;
            });
            data.regions.forEach(function (region) {
                region.cell = undefined;
                data.forEach(function (assignment) {
                    if (assignment.region === region.name) {
                        region.cell = data.grid[Number(assignment.gridX) + offX][Number(assignment.gridY) + offY];
                        if (!hasValue(region.cell)) {
                            debug.appendLine("Couldn't find cell: " + assignment.gridX + " " + assignment.gridY);
                        } else {
                            region.cell.region = region;
                        }
                    }
                });
                if (!hasValue(region.cell)) {
                    debug.appendLine("Couldn't find region: " + region.name);
                }
            });

            if (!keepgrid) {
                analyzeMap(true);
                analyzeGrid(true);
                analyzeMapAndGrid();
            }

            measureCurrentAssignment(undefined, undefined, true);
            visualizeAssignment();

            if (!keepgrid) {
                debug.appendLine("");
                appendMapGeometry(debug);
                debug.appendLine("");
                appendGridGeometry(debug);
            }

            batchLikeOutput = false;

            if (hasValue(callback)) {
                callback();
            }
        });
    }


    return dt;
})();