/*
    © Copyright 2016 Wouter Meulemans 
    
    This file is part of SMWGComputation.

    SMWGComputation is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SMWGComputation is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with SMWGComputation.  If not, see <http://www.gnu.org/licenses/>.
*/

/* global data */

var rectMath = {
    getPrimeShape: function () {
        // LT, RT, RB, LB
        return [
            // LT
            [0, 0],
            // RT
            [1, 0],
            // RB
            [1, 1],
            // LB
            [0, 1]
        ];
    },
    getCellGeometry: function (gridX, gridY, gridbase) {
        var result = this.getPrimeShape();
        result.forEach(function (corner) {
            corner[0] = gridbase.offsetX + (corner[0] + gridX) * gridbase.sizeX;
            corner[1] = gridbase.offsetY + (corner[1] + gridY) * gridbase.sizeY;
        });
        return result;
    },
    getCommonGeometry: function (cellA, cellB) {
        var dx = cellA.column - cellB.column;
        var dy = cellA.row - cellB.row;
        switch (dx) {
            case 0:
                // same column
                switch (dy) {
                    case 0:
                        // A == B
                        return cellA.geometry;
                    case 1:
                        // B
                        // A                        
                        return [cellA.geometry[2], cellA.geometry[3]];
                    case -1:
                        // A
                        // B
                        return [cellA.geometry[0], cellA.geometry[1]];
                }
                break;
            case 1:
                // A is right of B
                switch (dy) {
                    case 0:
                        // B A
                        return [cellA.geometry[3], cellA.geometry[0]];
                    case 1:
                        // B
                        //   A
                        return [cellA.geometry[0]];
                    case -1:
                        //   A
                        // B
                        return [cellA.geometry[3]];
                }
                break;
            case -1:
                // A is left of B
                switch (dy) {
                    case 0:
                        // A B
                        return [cellA.geometry[1], cellA.geometry[2]];
                    case 1:
                        //   B
                        // A
                        return [cellA.geometry[1]];
                    case -1:
                        // A
                        //   B
                        return [cellA.geometry[2]];
                }
                break;
        }

        return [];
    },
    getNeighbors: function (cell, onesided) {
        var nbs = [];
        var nb;
        // above
        nb = data.getGridCell(cell.column, cell.row - 1);
        if (nb !== null) {
            nbs.push(nb);
        }
        // above/left
        nb = data.getGridCell(cell.column - 1, cell.row - 1);
        if (nb !== null) {
            nbs.push(nb);
        }
        // left
        nb = data.getGridCell(cell.column - 1, cell.row);
        if (nb !== null) {
            nbs.push(nb);
        }
        // below/left
        nb = data.getGridCell(cell.column - 1, cell.row + 1);
        if (nb !== null) {
            nbs.push(nb);
        }
        if (!onesided) {
            // below
            nb = data.getGridCell(cell.column, cell.row + 1);
            if (nb !== null) {
                nbs.push(nb);
            }
            // below/right
            nb = data.getGridCell(cell.column + 1, cell.row + 1);
            if (nb !== null) {
                nbs.push(nb);
            }
            // right            
            nb = data.getGridCell(cell.column + 1, cell.row);
            if (nb !== null) {
                nbs.push(nb);
            }
            // above/right                   
            nb = data.getGridCell(cell.column + 1, cell.row - 1);
            if (nb !== null) {
                nbs.push(nb);
            }
        }
        return nbs;
    },
    hopDistance: function (cellA, cellB) {
        var dx = Math.abs(cellA.column - cellB.column);
        var dy = Math.abs(cellA.row - cellB.row);
        return Math.max(dx, dy);
    },
    computeSpanningBase: function (totwidth, totheight, ncolumns, nrows, fixratio) {
        var base = {};
        if (fixratio) {
            base.sizeX = base.sizeY = Math.min(totwidth / ncolumns, totheight / nrows);
        } else {
            base.sizeX = totwidth / ncolumns;
            base.sizeY = totheight / nrows;
        }
        base.offsetX = (500 - base.sizeX * ncolumns) / 2;
        base.offsetY = (500 - base.sizeY * nrows) / 2;
        return base;
    },
    computeAreaBase: function (ncolumns, nrows, cellarea, cellratio) {
        var base = {};
        base.sizeY = Math.sqrt(cellarea / cellratio);
        base.sizeX = base.sizeY * cellratio;
        base.offsetX = (500 - base.sizeX * ncolumns) / 2;
        base.offsetY = (500 - base.sizeY * nrows) / 2;
        return base;
    }
};

var hexHorzMath = {
    getPrimeShape: function () {
        return [
            // L
            [0, 0],
            // LT
            [0.288675, -0.5],
            // RT
            [0.866025, -0.5],
            // R
            [1.1547, 0],
            // RB
            [0.866025, 0.5],
            // LB
            [0.288675, 0.5]
        ];
    },
    getCellGeometry: function (gridX, gridY, gridbase) {
        var result = this.getPrimeShape();
        result.forEach(function (corner) {
            corner[0] = gridbase.offsetX + (corner[0] * gridbase.sizeX) + gridX * 0.866025 * gridbase.sizeX;
            corner[1] = gridbase.offsetY + ((corner[1] + 0.5) * gridbase.sizeY) + (gridY + (gridX % 2 === 0 ? 0 : 0.5)) * gridbase.sizeY;
        });
        return result;
    },
    getCommonGeometry: function (cellA, cellB) {
        var rowoff = cellA.column % 2 === 0 ? 1 : 0;
        if (cellA === cellB) {
            return cellA.geometry;
        } else
        if (cellB === data.getGridCell(cellA.column, cellA.row - 1)) {
            // B above A
            return [cellA.geometry[1], cellA.geometry[2]];
        } else if (cellB === data.getGridCell(cellA.column + 1, cellA.row - rowoff)) {
            // B above/right A
            return [cellA.geometry[2], cellA.geometry[3]];
        } else if (cellB === data.getGridCell(cellA.column + 1, cellA.row + 1 - rowoff)) {
            // B below/right A
            return [cellA.geometry[3], cellA.geometry[4]];
        } else if (cellB === data.getGridCell(cellA.column, cellA.row + 1)) {
            // B below A
            return [cellA.geometry[4], cellA.geometry[5]];
        } else if (cellB === data.getGridCell(cellA.column - 1, cellA.row + 1 - rowoff)) {
            // B below/left A
            return [cellA.geometry[5], cellA.geometry[0]];
        } else if (cellB === data.getGridCell(cellA.column - 1, cellA.row - rowoff)) {
            // B above/left A
            return [cellA.geometry[0], cellA.geometry[1]];
        } else {
            // nonadjacent
            return [];
        }
    },
    getNeighbors: function (cell, onesided) {
        var nbs = [];
        var rowoff = cell.column % 2 === 0 ? 1 : 0;
        var nb;
        // above
        nb = data.getGridCell(cell.column, cell.row - 1);
        if (nb !== null) {
            nbs.push(nb);
        }
        // above/right
        nb = data.getGridCell(cell.column + 1, cell.row - rowoff);
        if (nb !== null) {
            nbs.push(nb);
        }
        // below/right
        nb = data.getGridCell(cell.column + 1, cell.row + 1 - rowoff);
        if (nb !== null) {
            nbs.push(nb);
        }
        if (!onesided) {
            // below      
            nb = data.getGridCell(cell.column, cell.row + 1);
            if (nb !== null) {
                nbs.push(nb);
            }
            // below/left
            nb = data.getGridCell(cell.column - 1, cell.row + 1 - rowoff);
            if (nb !== null) {
                nbs.push(nb);
            }
            // above/left              
            nb = data.getGridCell(cell.column - 1, cell.row - rowoff);
            if (nb !== null) {
                nbs.push(nb);
            }
        }
        return nbs;
    },
    hopDistance: function (cellA, cellB) {
        var dx = Math.abs(cellA.column - cellB.column);
        var goDown, goUp;
        if (dx > 0.001) {
            goDown = goUp = cellA.row;
        } else if (cellA.row % 2 === 0) {
            goDown = cellA.row + Math.floor(0.5 * dx);
            goUp = cellA.row - Math.ceil(0.5 * dx);
        } else {
            goDown = cellA.row + Math.ceil(0.5 * dx);
            goUp = cellA.row - Math.floor(0.5 * dx);
        }
        if (cellB.row <= goUp) {
            return dx + (goUp - cellB.row);
        } else if (cellB.row >= goDown) {
            return dx + (cellB.row - goDown);
        } else {
            return dx;
        }
    },
    computeSpanningBase: function (totwidth, totheight, ncolumns, nrows, fixratio) {
        var base = {};
        // height of a hex row is 1.5, each row adds 1
        // totheight = sizeY * 0.5 + nrows * sizeY * 1
        // totheight = sizeY * (0.5 + nrows)
        // sizeY = totheight / (0.5 + nrows)

        // width of a hex is 1.1547, each col adds (1.1547 - 0.288675)
        // totwidth =  sizeX * 0.288675 + ncolumns * sizeX * (1.1547 - 0.288675)
        // totwidth = sizeX * (0.288675 + ncolumns * (1.1547 - 0.288675))
        // sizeX = totwidth / (0.288675 + ncolumns * (1.1547 - 0.288675))
        if (fixratio) {
            base.sizeX = base.sizeY = Math.min(
                    totwidth / (0.288675 + ncolumns * (1.1547 - 0.288675)),
                    totheight / (0.5 + nrows));
        } else {
            base.sizeX = totwidth / (0.288675 + ncolumns * (1.1547 - 0.288675));
            base.sizeY = totheight / (0.5 + nrows);
        }
        base.offsetX = (500 - base.sizeX * (0.288675 + ncolumns * (1.1547 - 0.288675))) / 2;
        base.offsetY = (500 - base.sizeY * (0.5 + nrows)) / 2;
        return base;
    },
    computeAreaBase: function (ncolumns, nrows, cellarea, cellratio) {
        var base = {};
        var primearea = computeArea(this.getPrimeShape());
        base.sizeY = Math.sqrt(cellarea / (cellratio * primearea));
        base.sizeX = base.sizeY * cellratio;
        base.offsetX = (500 - base.sizeX * (0.288675 + ncolumns * (1.1547 - 0.288675))) / 2;
        base.offsetY = (500 - base.sizeY * (0.5 + nrows)) / 2;
        return base;
    }
};

var hexVertMath = {
    getPrimeShape: function () {
        return [
            // T
            [0, 0],
            // RT
            [0.5, 0.288675],
            // RB
            [0.5, 0.866025],
            // B
            [0, 1.1547],
            // LB
            [-0.5, 0.866025],
            // LT
            [-0.5, 0.288675]
        ];
    },
    getCellGeometry: function (gridX, gridY, gridbase) {
        var result = this.getPrimeShape();
        result.forEach(function (corner) {
            corner[0] = gridbase.offsetX + ((corner[0] + 0.5) * gridbase.sizeX) + (gridX + (gridY % 2 === 0 ? 0 : 0.5)) * gridbase.sizeX;
            corner[1] = gridbase.offsetY + (corner[1] * gridbase.sizeY) + gridY * 0.866025 * gridbase.sizeY;
        });
        return result;
    },
    getCommonGeometry: function (cellA, cellB) {
        var coloff = cellA.row % 2 === 0 ? 1 : 0;
        if (cellA === cellB) {
            return cellA.geometry;
        } else if (cellB === data.getGridCell(cellA.column - coloff, cellA.row - 1)) {
            // B above/left A
            return [cellA.geometry[5], cellA.geometry[0]];
        } else if (cellB === data.getGridCell(cellA.column + 1 - coloff, cellA.row - 1)) {
            // B above/right A
            return [cellA.geometry[0], cellA.geometry[1]];
        } else if (cellB === data.getGridCell(cellA.column + 1, cellA.row)) {
            // B right A
            return [cellA.geometry[1], cellA.geometry[2]];
        } else if (cellB === data.getGridCell(cellA.column + 1 - coloff, cellA.row + 1)) {
            // B below/right A   
            return [cellA.geometry[2], cellA.geometry[3]];
        } else if (cellB === data.getGridCell(cellA.column - coloff, cellA.row + 1)) {
            // B below/left A
            return [cellA.geometry[3], cellA.geometry[4]];
        } else if (cellB === data.getGridCell(cellA.column - 1, cellA.row)) {
            // B left A 
            return [cellA.geometry[4], cellA.geometry[5]];
        } else {
            return [];
        }
    },
    getNeighbors: function (cell, onesided) {
        var nbs = [];
        var coloff = cell.row % 2 === 0 ? 1 : 0;
        var nb;
        // above/left
        nb = data.getGridCell(cell.column - coloff, cell.row - 1);
        if (nb !== null) {
            nbs.push(nb);
        }
        // above/right
        nb = data.getGridCell(cell.column + 1 - coloff, cell.row - 1);
        if (nb !== null) {
            nbs.push(nb);
        }
        // right
        nb = data.getGridCell(cell.column + 1, cell.row);
        if (nb !== null) {
            nbs.push(nb);
        }
        if (!onesided) {
            // below/right    
            nb = data.getGridCell(cell.column + 1 - coloff, cell.row + 1);
            if (nb !== null) {
                nbs.push(nb);
            }
            // below/left
            nb = data.getGridCell(cell.column - coloff, cell.row + 1);
            if (nb !== null) {
                nbs.push(nb);
            }
            // left              
            nb = data.getGridCell(cell.column - 1, cell.row);
            if (nb !== null) {
                nbs.push(nb);
            }
        }
        return nbs;
    },
    hopDistance: function (cellA, cellB) {
        var dy = Math.abs(cellA.row - cellB.row);
        var goRight, goLeft;
        if (dy > 0.0001) {
            goRight = goLeft = cellA.column;
        } else if (cellA.column % 2 === 0) {
            goRight = cellA.column + Math.floor(0.5 * dy);
            goLeft = cellA.column - Math.ceil(0.5 * dy);
        } else {
            goRight = cellA.column + Math.ceil(0.5 * dy);
            goLeft = cellA.column - Math.floor(0.5 * dy);
        }
        if (cellB.column <= goLeft) {
            return dy + (goLeft - cellB.column);
        } else if (cellB.column >= goRight) {
            return dy + (cellB.column - goRight);
        } else {
            return dy;
        }
    },
    computeSpanningBase: function (totwidth, totheight, ncolumns, nrows, fixratio) {
        var base = {};
        // inverse of hexHorzMath
        if (fixratio) {
            base.sizeX = base.sizeY = Math.min(
                    totwidth / (0.5 + ncolumns),
                    totheight / (0.288675 + nrows * (1.1547 - 0.288675)));
        } else {
            base.sizeX = totwidth / (0.5 + ncolumns);
            base.sizeY = totheight / (0.288675 + nrows * (1.1547 - 0.288675));
        }
        base.offsetX = (500 - base.sizeX * (0.5 + ncolumns)) / 2;
        base.offsetY = (500 - base.sizeY * (0.288675 + nrows * (1.1547 - 0.288675))) / 2;
        return base;
    },
    computeAreaBase: function (ncolumns, nrows, cellarea, cellratio) {
        var base = {};
        var primearea = computeArea(this.getPrimeShape());
        base.sizeY = Math.sqrt(cellarea / (cellratio * primearea));
        base.sizeX = base.sizeY * cellratio;
        base.offsetX = (500 - base.sizeX * (0.5 + ncolumns)) / 2;
        base.offsetY = (500 - base.sizeY * (0.288675 + nrows * (1.1547 - 0.288675))) / 2;
        return base;
    }
};