/*
    © Copyright 2016 Wouter Meulemans 
    
    This file is part of SMWGComputation.

    SMWGComputation is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SMWGComputation is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with SMWGComputation.  If not, see <http://www.gnu.org/licenses/>.
*/

/* global data, debug, visual, batch */

var lp = (function () {

    var lpo = {
        compute: _compute,
        shuffle: _shuffle,
        shuffleGeodesic: _geodesic
    };

    function _compute(omitEmptyCells, state, suppressreport, dfunc) {
        if (!hasValue(data.grid)) {
            return;
        }
        if (!hasValue(data.regions)) {
            return;
        }
        if (data.regions.length > data.gridcells.length) {
            return;
        }

        var cellsToWorkWith;

        if (omitEmptyCells) {
            cellsToWorkWith = [];
            data.regions.forEach(function (region) {
                if (state) {
                    cellsToWorkWith.push(state.region_to_grid[region.index]);
                } else if (hasValue(region.cell)) {
                    cellsToWorkWith.push(region.cell);
                }
            });
            if (cellsToWorkWith.length !== data.regions.length) {
                return;
            }
        } else {
            cellsToWorkWith = data.gridcells;
        }

        if (!hasValue(dfunc)) {
            dfunc = function (region, cell) {
                // squared euclidean
                var dx = cell.center[0] - region.center[0];
                var dy = cell.center[1] - region.center[1];
                return dx * dx + dy * dy;
            };
        }

        var lp_start = performance.now();

        var solver = window.solver;
        var model = {
            optimize: "cost",
            opType: "min",
            constraints: {},
            variables: {}
        };
        var nregions = data.regions.length;
        var ncells = cellsToWorkWith.length;

        // constraints
        // at most region per cell
        for (var cellindex = 0; cellindex < ncells; cellindex++) {
            model.constraints["c" + cellindex] = {max: 1};
        }

        // (at least) one cell per region
        for (var regionindex = 0; regionindex < nregions; regionindex++) {
            model.constraints["r" + regionindex] = {min: 1};
        }

        // variables
        // for each region-cell pair, make a variable
        for (var regionindex = 0; regionindex < nregions; regionindex++) {
            for (var cellindex = 0; cellindex < ncells; cellindex++) {

                var dist_from_func = dfunc(data.regions[regionindex], cellsToWorkWith[cellindex]);
                if (!Number.isFinite(dist_from_func)) {
                    //console.log("Infinite : " + dist_from_func + " = d(" + regions[regionindex].name + ", [" + cellsToWorkWith[cellindex].column + "," + cellsToWorkWith[cellindex].row + "])");
                    dist_from_func = 5000 * 5000;
                    // just use some big number to avoid the LP from not functioning
                }
                var varobj = {
                    cost: dist_from_func
                };
                varobj["r" + regionindex] = 1;
                varobj["c" + cellindex] = 1;
                model.variables["r" + regionindex + "c" + cellindex] = varobj;
            }
        }

        var lp_init = performance.now();
        var results = solver.Solve(model);
        var lp_end = performance.now();

        if (!suppressreport) {
            if (batch.active) {
                batch.result_data.append("\t" + (lp_end - lp_start));
            } else if (state) {
                if (omitEmptyCells) {
                    debug.appendLine("LP SHUFFLE");
                } else {
                    debug.appendLine("LP ASSIGNMENT");
                }
                debug.appendLine("in " + (lp_end - lp_start) + "ms");
            } else {
                debug.clear();
                if (omitEmptyCells) {
                    debug.appendLine("LP SHUFFLE under L_2^2 displacement");
                } else {
                    debug.appendLine("LP ASSIGNMENT under L_2^2 displacement");
                }
                debug.appendLine("init time: \t" + (lp_init - lp_start) + "ms");
                debug.appendLine("solve time: \t" + (lp_end - lp_init) + "ms");
                debug.appendLine("total time: \t" + (lp_end - lp_start) + "ms");
            }
        }

        if (state) {
            cellsToWorkWith.forEach(function (cell) {
                state.grid_to_region[cell.index] = null;
            });
        } else {
            data.regions.forEach(function (region) {
                region.cell = undefined;
            });
            data.gridcells.forEach(function (gridcell) {
                gridcell.region = undefined;
            });
        }

        for (var regionindex = 0; regionindex < nregions; regionindex++) {
            for (var cellindex = 0; cellindex < ncells; cellindex++) {
                if (results["r" + regionindex + "c" + cellindex] > 0.5) {
                    var cell = cellsToWorkWith[cellindex];
                    var region = data.regions[regionindex];
                    if (!state) {
                        cell.region = region;
                        region.cell = cell;
                    } else {
                        state.grid_to_region[cell.index] = region;
                        state.region_to_grid[region.index] = cell;
                    }
                }
            }
        }

        if (!state) {
            outputAssignment();
            visual.visualizeAssignment();
        }
    }

    function _shuffle(state, suppressreport) {
        _compute(true, state, suppressreport);
    }

    function _geodesic() {

        var addingdebuggeometry = false;
        var addblocked = false;
        var onlybaseblocked = true;

        debug.clear();
        var ggrid_debug;
        if (addingdebuggeometry) {
            visual.ggrid_transform.selectAll("g.debug").remove();
            ggrid_debug = visual.ggrid_transform.append("g").attr("class", "debug");
        }

        var ncolumns = data.grid.length;
        var nrows = data.grid[0].length;

        function isoccuppied(col, row) {
            var cell = data.getGridCell(col, row);
            if (cell === null) {
                return false;
            } else {
                return hasValue(cell.region);
            }
        }

        // NB: returns false if cell doesnt exist!
        function isempty(col, row) {
            var cell = data.getGridCell(col, row);
            if (cell === null) {
                return false;
            } else {
                return !hasValue(cell);
            }
        }

        var basestate = [];
        for (var i = 0; i < data.gridcells.length; i++) {
            // NB: cell is LT,RT,RB,LB
            // NB: only have to add reflex vertices
            var cell = data.gridcells[i];
            if (!hasValue(cell.region)) {
                continue;
            }
            // 0 - LT:
            // occ occ
            // emp cel
            if (isoccuppied(cell.column, cell.row - 1) && isoccuppied(cell.column - 1, cell.row - 1) && isempty(cell.column - 1, cell.row)) {
                basestate.push({
                    point: cell.geometry[0],
                    forbidden_1: [-1, 0],
                    forbidden_2: [0, 1]
                });
            }
            // 1 - RT:
            // emp occ
            // cel occ
            if (isoccuppied(cell.column + 1, cell.row) && isoccuppied(cell.column + 1, cell.row - 1) && isempty(cell.column, cell.row - 1)) {
                basestate.push({
                    point: cell.geometry[1],
                    forbidden_1: [-1, 0],
                    forbidden_2: [0, -1]
                });
            }
            // 2 - RB:
            // cel emp
            // occ occ
            if (isoccuppied(cell.column, cell.row + 1) && isoccuppied(cell.column + 1, cell.row + 1) && isempty(cell.column + 1, cell.row)) {
                basestate.push({
                    point: cell.geometry[2],
                    forbidden_1: [1, 0],
                    forbidden_2: [0, -1]
                });
            }
            // 3 - LB:
            // occ cel
            // occ emp
            if (isoccuppied(cell.column - 1, cell.row) && isoccuppied(cell.column - 1, cell.row + 1) && isempty(cell.column, cell.row + 1)) {
                basestate.push({
                    point: cell.geometry[3],
                    forbidden_1: [1, 0],
                    forbidden_2: [0, 1]
                });
            }
        }

        function test_horizontal_below(col, row) {
            return (isoccuppied(col, row) && isempty(col, row + 1))
                    || (isoccuppied(col, row + 1) && isempty(col, row));
        }

        function test_vertical_right(col, row) {
            return (isoccuppied(col, row) && isempty(col + 1, row))
                    || (isoccuppied(col + 1, row) && isempty(col, row));
        }

        var verticals = [];
        var horizontals = [];
        for (var col = 0; col < ncolumns; col++) {
            for (var row = 0; row < nrows; row++) {
                // NB: cell is LT,RT,RB,LB
                var cell = data.grid[col][row];
                if (row < nrows - 1 &&
                        test_horizontal_below(col, row) &&
                        !test_horizontal_below(col - 1, row)) {
                    horizontals.push({
                        y: cell.geometry[2][1],
                        xleft: cell.geometry[3][0],
                        xrght: cell.geometry[2][0]
                    });
                    var walkright = col + 1;
                    while (test_horizontal_below(walkright, row)) {
                        horizontals[horizontals.length - 1].xrght = data.grid[walkright][row].geometry[2][0];
                        walkright++;
                    }

                    // extend into void?
                    if (!isoccuppied(col - 1, row) && !isoccuppied(col - 1, row + 1)) {
                        horizontals[horizontals.length - 1].xleft -= 0.1 * (cell.geometry[1][0] - cell.geometry[0][0]);
                    }
                    if (!isoccuppied(walkright, row) && !isoccuppied(walkright, row + 1)) {
                        horizontals[horizontals.length - 1].xrght += 0.1 * (cell.geometry[1][0] - cell.geometry[0][0]);
                    }
                }

                if (col < ncolumns - 1 &&
                        test_vertical_right(col, row) &&
                        !test_vertical_right(col, row - 1)) {
                    verticals.push({
                        x: cell.geometry[1][0],
                        ytop: cell.geometry[1][1],
                        ybot: cell.geometry[2][1]
                    });
                    var walkdown = row + 1;
                    while (test_vertical_right(col, walkdown)) {
                        verticals[verticals.length - 1].ybot = data.grid[col][walkdown].geometry[2][1];
                        walkdown++;
                    }

                    // extend into void?
                    if (!isoccuppied(col, row - 1) && !isoccuppied(col + 1, row - 1)) {
                        verticals[verticals.length - 1].ytop -= 0.1 * (cell.geometry[2][1] - cell.geometry[1][1]);
                    }
                    if (!isoccuppied(col, walkdown) && !isoccuppied(col + 1, walkdown)) {
                        verticals[verticals.length - 1].ybot += 0.1 * (cell.geometry[2][1] - cell.geometry[1][1]);
                    }
                }
            }
        }

        function lineintersecthorz(p0_x, p0_y, p1_x, p1_y, xleft, xrght, y) {

            var maxy = Math.max(p0_y, p1_y);
            var miny = Math.min(p0_y, p1_y);
            if (y >= maxy - 0.0001) {
                return false;
            } else if (y <= miny + 0.0001) {
                return false;
            } else {
                var x;
                if (p0_y <= p1_y) {
                    x = p0_x + (p1_x - p0_x) * (y - miny) / (maxy - miny);
                } else {
                    x = p1_x + (p0_x - p1_x) * (y - miny) / (maxy - miny);
                }
                if (x >= xrght - 0.0001) {
                    return false;
                } else if (x <= xleft + 0.0001) {
                    return false;
                } else {
                    return true;
                }
            }
        }

        function lineintersectvert(p0_x, p0_y, p1_x, p1_y, x, ytop, ybot, vert) {
            var maxx = Math.max(p0_x, p1_x);
            var minx = Math.min(p0_x, p1_x);
            if (x >= maxx - 0.0001) {
                return false;
            } else if (x <= minx + 0.0001) {
                return false;
            } else {
                var y;
                if (p0_x <= p1_x) {
                    y = p0_y + (p1_y - p0_y) * (x - minx) / (maxx - minx);
                } else {
                    y = p1_y + (p0_y - p1_y) * (x - minx) / (maxx - minx);
                }
                if (y >= ybot - 0.0001) {
                    return false;
                } else if (y <= ytop + 0.0001) {
                    return false;
                } else {
                    return true;
                }
            }
        }

        var baseindex = [];
        var cellvisindex = [];
        basestate.forEach(function (st) {
            var list = [];
            baseindex.push(list);
            basestate.forEach(function (st2) {
                list.push(0);
            });

            var list2 = [];
            cellvisindex.push(list2);
            data.gridcells.forEach(function (c) {
                list2.push(0);
            });
        });

        function isvisible(st_from, st_to, targetcell) {
            var fi = st_from.baseindex;
            var ti = st_to.baseindex;
            var vis;
            var newtestresult;
            if (fi >= 0 && ti >= 0) {
                // between to base points
                if (baseindex[fi][ti] === 0) {
                    newtestresult = testvisible(st_from, st_to);
                    if (newtestresult === 1) {
                        baseindex[fi][ti] = 1;
                        baseindex[ti][fi] = 1;
                    } else {
                        baseindex[fi][ti] = -1;
                        baseindex[ti][fi] = -1;
                    }
                }
                vis = baseindex[fi][ti] === 1;
            } else if (fi >= 0) {
                ti = targetcell.index;
                // between basepoint and target cell center
                if (cellvisindex[fi][ti] === 0) {
                    newtestresult = testvisible(st_from, st_to);
                    if (newtestresult === 1) {
                        cellvisindex[fi][ti] = 1;
                    } else {
                        cellvisindex[fi][ti] = -1;
                    }
                }
                vis = cellvisindex[fi][ti] === 1;
            } else if (ti >= 0) {
                // between region centroid and base
                newtestresult = testvisible(st_from, st_to);
                vis = newtestresult === 1;
            } else {
                // between region centroid and cell centriod  
                newtestresult = testvisible(st_from, st_to);
                vis = newtestresult === 1;
            }

            if (newtestresult !== undefined && addingdebuggeometry) {
                if (addblocked && (!onlybaseblocked || (st_from.baseindex >= 0 && st_to.baseindex >= 0)) || newtestresult === 1) {
                    var testcolor;
                    switch (newtestresult) {
                        case 1:
                            testcolor = "green";
                            break;
                        case 2:
                            testcolor = "red";
                            break;
                        case 3:
                            testcolor = "orange";
                            break;
                        case 4:
                            testcolor = "pink";
                            break;
                        default:
                            testcolor = "black";
                            break;
                    }
                    ggrid_debug.append("path")
                            .attr("class", "pathway")
                            .attr("style", "stroke: " + testcolor + "; stroke-width: 1; fill: none;")
                            .attr("d", "M" + st_from.point[0] + "," + st_from.point[1] + "L" + st_to.point[0] + "," + st_to.point[1]);
                }
            }

            return vis;
        }

        function testvisible(st_from, st_to) {
            var dx = st_to.point[0] - st_from.point[0];
            var dy = st_to.point[1] - st_from.point[1];

            // check if the forbidden directions are exactly one satisfied
            if (st_from.baseindex >= 0) {
                if (0.000001 < (dx * basestate[st_from.baseindex].forbidden_1[0] + dy * basestate[st_from.baseindex].forbidden_1[1])
                        * (dx * basestate[st_from.baseindex].forbidden_2[0] + dy * basestate[st_from.baseindex].forbidden_2[1])) {
                    return 3;
                }
            }
            if (st_to.baseindex >= 0) {
                if (0.000001 < (dx * basestate[st_to.baseindex].forbidden_1[0] + dy * basestate[st_to.baseindex].forbidden_1[1])
                        * (dx * basestate[st_to.baseindex].forbidden_2[0] + dy * basestate[st_to.baseindex].forbidden_2[1])) {
                    return 3;
                }
            }

            var dist = Math.sqrt(dx * dx + dy * dy);
            for (var bsi = 0; bsi < basestate.length; bsi++) {
                if (bsi === st_from.baseindex || bsi === st_to.baseindex) {
                    continue;
                }

                var dxbs = basestate[bsi].point[0] - st_from.point[0];
                var dybs = basestate[bsi].point[1] - st_from.point[1];
                var distbs = Math.sqrt(dxbs * dxbs + dybs * dybs);

                if (0 < distbs && distbs < dist && dx * dxbs + dy * dybs >= dist * distbs * 0.99999) {
                    return 4;
                }
            }

            if (Math.abs(dx) < 0.0001) {
                // no difference in x, so vertical line
                var x = st_from.point[0];
                var ytop = Math.min(st_from.point[1], st_to.point[1]);
                var ybot = Math.max(st_from.point[1], st_to.point[1]);
                for (var h = 0; h < horizontals.length; h++) {
                    var horz = horizontals[h];
                    if (horz.xleft + 0.0001 <= x && x <= horz.xrght - 0.0001) {
                        if (ytop <= horz.y && horz.y <= ybot) {
                            return 2;
                        }
                    }
                }
            } else if (Math.abs(dy) < 0.0001) {
                // no difference in y, so horizontal line
                var y = st_from.point[1];
                var xleft = Math.min(st_from.point[0], st_to.point[0]);
                var xrght = Math.max(st_from.point[0], st_to.point[0]);
                for (var v = 0; v < verticals.length; v++) {
                    var vert = verticals[v];
                    if (vert.ytop + 0.0001 <= y && y <= vert.ybot - 0.0001) {
                        if (xleft <= vert.x && vert.x <= xrght) {
                            return 2;
                        }
                    }
                }
            } else {
                // general line
                for (var h = 0; h < horizontals.length; h++) {
                    var horz = horizontals[h];
                    if (lineintersecthorz(st_from.point[0], st_from.point[1],
                            st_to.point[0], st_to.point[1],
                            horz.xleft, horz.xrght, horz.y)) {
                        return 2;
                    }
                }

                for (var v = 0; v < verticals.length; v++) {
                    var vert = verticals[v];
                    if (lineintersectvert(st_from.point[0], st_from.point[1],
                            st_to.point[0], st_to.point[1],
                            vert.x, vert.ytop, vert.ybot)) {
                        return 2;
                    }
                }
            }

            return 1;
        }

        var regionstarts = [];
        data.regions.forEach(function (region) {

            var startpoint;
            var startpointdistance = Number.POSITIVE_INFINITY;
            // check if region center lies in a grid cell, if so, just use that, otherwise, find the nearest one!

            for (var ri = 0; ri < data.regions.length; ri++) {
                var cell = data.gridcells[data.regions[ri].cell.index];
                // NB: geom = LT, RT, RB, LB
                if (region.center[0] <= cell.geometry[0][0]) {
                    // left of cell      
                    if (region.center[1] <= cell.geometry[0][1]) {
                        // above cell 
                        var cand = cell.geometry[0];
                        var d = computeEuclideanDistance(cand, region.center);
                        if (d < startpointdistance) {
                            startpoint = cand;
                            startpointdistance = d;
                        }
                    } else if (region.center[1] >= cell.geometry[2][1]) {
                        // below cell
                        var cand = cell.geometry[3];
                        var d = computeEuclideanDistance(cand, region.center);
                        if (d < startpointdistance) {
                            startpoint = cand;
                            startpointdistance = d;
                        }
                    } else {
                        // in row of cell
                        var cand = [cell.geometry[0][0], region.center[1]];
                        var d = cand[0] - region.center[0];
                        if (d < startpointdistance) {
                            startpoint = cand;
                            startpointdistance = d;
                        }
                    }
                } else if (region.center[0] >= cell.geometry[1][0]) {
                    // right of cell
                    if (region.center[1] <= cell.geometry[0][1]) {
                        // above cell   
                        var cand = cell.geometry[1];
                        var d = computeEuclideanDistance(cand, region.center);
                        if (d < startpointdistance) {
                            startpoint = cand;
                            startpointdistance = d;
                        }
                    } else if (region.center[1] >= cell.geometry[2][1]) {
                        // below cell
                        var cand = cell.geometry[2];
                        var d = computeEuclideanDistance(cand, region.center);
                        if (d < startpointdistance) {
                            startpoint = cand;
                            startpointdistance = d;
                        }
                    } else {
                        // in row of cell
                        var cand = [cell.geometry[1][0], region.center[1]];
                        var d = region.center[0] - cand[0];
                        if (d < startpointdistance) {
                            startpoint = cand;
                            startpointdistance = d;
                        }
                    }
                } else {
                    // in column of cell
                    if (region.center[1] <= cell.geometry[0][1]) {
                        // above cell     
                        var cand = [region.center[0], cell.geometry[0][1]];
                        var d = cand[1] - region.center[1];
                        if (d < startpointdistance) {
                            startpoint = cand;
                            startpointdistance = d;
                        }
                    } else if (region.center[1] >= cell.geometry[2][1]) {
                        // below cell
                        var cand = [region.center[0], cell.geometry[2][1]];
                        var d = region.center[1] - cand[1];
                        if (d < startpointdistance) {
                            startpoint = cand;
                            startpointdistance = d;
                        }
                    } else {
                        // in row of cell
                        startpoint = region.center;
                        startpointdistance = 0;
                    }
                }
            }

            regionstarts.push({
                point: startpoint,
                distance: startpointdistance
            });

            if (addingdebuggeometry && startpointdistance > 0) {
                ggrid_debug.append("path")
                        .attr("class", "pathway")
                        .attr("style", "stroke: green; stroke-width: 1; fill: none;")
                        .attr("d", "M" + region.center[0] + "," + region.center[1] + "L" + startpoint[0] + "," + startpoint[1]);
            }
        });


        var cnt = 1;
        var computeDist = function (region, targetcell) {
            var state = [];
            cnt++;

            basestate.forEach(function (st, i) {
                state.push({
                    point: st.point,
                    baseindex: i,
                    distance: Number.POSITIVE_INFINITY,
                    prev: null
                });
            });

            state.push({
                point: targetcell.center,
                baseindex: -1,
                distance: Number.POSITIVE_INFINITY,
                prev: null
            });

            state.push({
                point: regionstarts[region.index].point,
                baseindex: -2,
                distance: regionstarts[region.index].distance,
                prev: null
            });

            while (true) {

                var min = null;
                var minindex = -1;
                state.forEach(function (st, i) {
                    if (min === null || st.distance < min.distance) {
                        min = st;
                        minindex = i;
                    }
                });
                if (!Number.isFinite(min.distance) || min.baseindex === -1) {
//                    debug.appendLine(region.name + " -> " + targetcell.column + "," + targetcell.row + " : " + min.distance);
//                        ggrid_debug.append("path")
//                                .attr("class", "pathway")
//                                .attr("style", "stroke: purple; stroke-width: 3; fill: none;")
//                                .attr("d", function () {
//                                    var pathcommand = "M" + min.point[0] + "," + min.point[1] + "L";
//                                    var walk = min.prev;
//                                    while (walk != null) {
//                                        pathcommand += " " + walk.point[0] + "," + walk.point[1];
//                                        walk = walk.prev;
//                                    }
//                                    pathcommand += " " + region.center[0] + "," + region.center[1];
//                                    return pathcommand;
//                                });    
                    return min.distance * min.distance;
                } else {
                    state.splice(minindex, 1);

                    state.forEach(function (st) {
                        if (isvisible(min, st, targetcell)) {
                            var dist = min.distance + computeEuclideanDistance(min.point, st.point);
                            if (dist < st.distance) {
                                st.distance = dist;
                                st.prev = min;
                            }
                        }
                    });
                }
            }
        };

        if (addingdebuggeometry) {
            var verts = ggrid_debug.selectAll("path.vertical").data(verticals);
            verts.exit().remove();
            verts.enter().append("path")
                    .attr("class", "vertical")
                    .attr("style", "stroke: blue; stroke-width: 2; fill: none;")
                    .attr("d", function (vert) {
                        return "M" + vert.x + "," + vert.ytop + "L" + vert.x + "," + vert.ybot;
                    });

            var horzs = ggrid_debug.selectAll("path.horizontal").data(horizontals);
            horzs.exit().remove();
            horzs.enter().append("path")
                    .attr("class", "horizontal")
                    .attr("style", "stroke: yellow; stroke-width: 2; fill: none;")
                    .attr("d", function (horz) {
                        return "M" + horz.xleft + "," + horz.y + "L" + horz.xrght + "," + horz.y;
                    });

            var forbs = ggrid_debug.selectAll("path.forbidden").data(basestate);
            forbs.exit().remove();
            forbs.enter().append("path")
                    .attr("class", "forbidden")
                    .attr("style", "stroke: black; stroke-width: 2; fill: none;")
                    .attr("d", function (bs) {
                        return "M" + bs.point[0] + "," + bs.point[1] + "L" + (bs.point[0] + 10 * (bs.forbidden_1[0] + bs.forbidden_2[0])) + "," + (bs.point[1] + 10 * (bs.forbidden_1[1] + bs.forbidden_2[1]));
                    });
        }

        _compute(true, undefined, true, computeDist);
    }

    return lpo;
})();