/*
    © Copyright 2016 Wouter Meulemans 
    
    This file is part of SMWGComputation.

    SMWGComputation is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SMWGComputation is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with SMWGComputation.  If not, see <http://www.gnu.org/licenses/>.
*/

/* global d3 */
/* global data, batch, debug, batchLikeOutput */

function getSettings(basicOnly, getAll) {

    var settings = {};

    var ncells = basicOnly ? Number.NaN : data.gridcells.length;
    var nregions = basicOnly ? Number.NaN : data.regions.length;

    function find_hop_pairs(hops) {
        var pairs = [];
        data.regions.forEach(function (region) {
            var visited = [];
            for (var i = 0; i < nregions; i++) {
                visited[i] = false;
            }
            visited[region.index] = true;
            var queue = [{r: region, d: 0}];
            while (queue.length > 0) {
                var head = queue.shift();
                head.r.topology.forEach(function (tp) {
                    if (visited[tp.neighbor.index]) {
                        // skip
                    } else {
                        if (tp.neighbor.index > head.r.index) {
                            pairs.push([head.r, tp.neighbor]);
                        }
                        if (head.d < hops - 1) {
                            queue.push({r: tp.neighbor, d: head.d + 1});
                        }
                    }
                });
            }
        });
        return pairs;
    }


    var neighborpairs = basicOnly ? [] : find_hop_pairs(1);

    settings.sa = {};
    settings.sa.rounds = Number(d3.select("#sa-rounds").property("value"));
    settings.sa.tempscale = Number(d3.select("#sa-tempscale").property("value"));
    settings.sa.init = d3.select("#sa-init").property("value");
    settings.sa.hillclimbing = d3.select("#sa-hillclimbing").property("checked");
    settings.sa.hillclimbingtime = d3.select("#sa-hillclimbing-time").property("value");
    settings.sa.invaropt = d3.select("#sa-invaropt").property("checked");
    settings.sa.maintainbest = d3.select("#sa-maintainbest").property("checked");

    if (batch.active && batch.writeheaders) {
        batch.settings_data.appendLine("Rounds\t" + settings.sa.rounds);
        batch.settings_data.appendLine("Temp scale\t" + settings.sa.tempscale);
        batch.settings_data.appendLine("Initialization\t" + settings.sa.init);
        batch.settings_data.appendLine("Hillclimb enabled\t" + settings.sa.hillclimbing);
        batch.settings_data.appendLine("HC time limit\t" + settings.sa.hillclimbingtime);
        batch.settings_data.appendLine("Invariance opt\t" + settings.sa.invaropt);
        batch.settings_data.appendLine("Maintain best\t" + settings.sa.maintainbest);
    }

    
    // note that there is a difference between the naming used in code and those presented!
    settings.measures = [];
    settings.getMeasureByName = function (name) {
        for (var i = 0; i < this.measures.length; i++) {
            if (this.measures[i].name === name) {
                return this.measures[i];
            }
        }
        return null;
    };
    settings.getMeasureByInterfacename = function (name) {
        for (var i = 0; i < this.measures.length; i++) {
            if (this.measures[i].interfacename === name) {
                return this.measures[i];
            }
        }
        return null;
    };
    
    if (getAll || d3.select("#loc-indiv-enabled").property("checked")) {
        var max;
        if (basicOnly) {
            max = 0;
        } else {
            max = 0;
            data.regions.forEach(function (regionA) {
                var maxD = 0;
                data.gridcells.forEach(function (cellA) {
                    maxD = Math.max(maxD, computeSquaredEuclideanDistance(regionA.center, cellA.center));
                });
                max += maxD;
            });
        }
        settings.measures.push({
            name: "LOC INDIV",
            interfacename: "Displacement",
            label: "loc-indiv",
            weight: getAll ? 0 : Number(d3.select("#loc-indiv-weight").property("value")),
            maxmeasure: max,
            invariance: {
                translation: false,
                rotation: false,
                scale: false,
                regions: false
            },
            measure: function (state) {
                var msr = 0;
                for (var irA = 0; irA < nregions; irA++) {
                    var regionA = data.regions[irA];
                    var cellA = state.region_to_grid[irA];
                    msr += computeSquaredEuclideanDistance(regionA.center, cellA.center);
                }
                return msr;
            },
            sa_different: false,
            sa_measure: function (state) {
                return this.measure(state);
            }
        });
    }
   
    if (getAll || d3.select("#loc-allpairs-enabled").property("checked")) {
        var max;
        if (basicOnly) {
            max = 0;
        } else {
            var maxD = computeEuclideanDistance(data.gridcells[0].center, data.gridcells[data.gridcells.length - 1].center);
            max = 0;
            for (var irA = 0; irA < nregions; irA++) {
                var regionA = data.regions[irA];
                for (var irB = irA + 1; irB < nregions; irB++) {
                    var regionB = data.regions[irB];
                    var rD = computeEuclideanDistance(regionA.center, regionB.center);
                    max += Math.pow(rD + maxD, 2);
                }
            }
        }
        settings.measures.push({
            name: "LOC AP",
            interfacename: "Vector (all)",
            label: "loc-allpairs",
            weight: getAll ? 0 : Number(d3.select("#loc-allpairs-weight").property("value")),
            maxmeasure: max,
            invariance: {
                translation: true,
                rotation: false,
                scale: false,
                regions: false
            },
            measure: function (state) {
                var msr = 0;
                for (var irA = 0; irA < nregions; irA++) {
                    var regionA = data.regions[irA];
                    var cellA = state.region_to_grid[irA];
                    for (var irB = irA + 1; irB < nregions; irB++) {
                        var regionB = data.regions[irB];
                        var cellB = state.region_to_grid[irB];
                        var dR = [regionA.center[0] - regionB.center[0],
                            regionA.center[1] - regionB.center[1]];
                        var dC = [cellA.center[0] - cellB.center[0],
                            cellA.center[1] - cellB.center[1]];
                        msr += computeSquaredEuclideanDistance(dR, dC);
                    }
                }
                return msr;
            },
            sa_different: false,
            sa_measure: function (state) {
                return this.measure(state);
            }
        });
    }

    if (getAll || d3.select("#loc-neighbors-enabled").property("checked")) {
        var max;
        if (basicOnly) {
            max = 0;
        } else {
            var maxD = computeEuclideanDistance(data.gridcells[0].center, data.gridcells[data.gridcells.length - 1].center);
            max = 0;
            neighborpairs.forEach(function (pair) {
                var regionA = pair[0];
                var regionB = pair[1];
                var rD = computeEuclideanDistance(regionA.center, regionB.center);
                max += Math.pow(rD + maxD, 2);
            });
        }
        settings.measures.push({
            name: "LOC N",
            interfacename: "Vector (nbr)",
            label: "loc-neighbors",
            weight: getAll ? 0 : Number(d3.select("#loc-neighbors-weight").property("value")),
            maxmeasure: max,
            invariance: {
                translation: true,
                rotation: false,
                scale: false,
                regions: false
            },
            measure: function (state) {
                var msr = 0;
                neighborpairs.forEach(function (pair) {
                    var regionA = pair[0];
                    var cellA = state.region_to_grid[regionA.index];
                    var regionB = pair[1];
                    var cellB = state.region_to_grid[regionB.index];
                    var dR = [regionA.center[0] - regionB.center[0],
                        regionA.center[1] - regionB.center[1]];
                    var dC = [cellA.center[0] - cellB.center[0],
                        cellA.center[1] - cellB.center[1]];
                    msr += computeSquaredEuclideanDistance(dR, dC);
                });
                return msr;
            },
            sa_different: false,
            sa_measure: function (state) {
                return this.measure(state);
            }
        });
    }

    if (getAll || d3.select("#dist-allpairs-enabled").property("checked")) {
        var max;
        if (basicOnly) {
            max = 0;
        } else {
            var minD = computeEuclideanDistance(data.gridcells[0].center, data.gridcells[1].center);
            var maxD = computeEuclideanDistance(data.gridcells[0].center, data.gridcells[data.gridcells.length - 1].center);
            max = 0;
            for (var irA = 0; irA < nregions; irA++) {
                var regionA = data.regions[irA];
                for (var irB = irA + 1; irB < nregions; irB++) {
                    var regionB = data.regions[irB];
                    var rD = computeEuclideanDistance(regionA.center, regionB.center);
                    max += Math.max(
                            Math.pow(rD - minD, 2),
                            Math.pow(rD - maxD, 2));
                }
            }
        }
        settings.measures.push({
            name: "DIST AP",
            interfacename: "Distance (all)",
            label: "dist-allpairs",
            weight: getAll ? 0 : Number(d3.select("#dist-allpairs-weight").property("value")),
            maxmeasure: max,
            invariance: {
                translation: true,
                rotation: true,
                scale: false,
                regions: false
            },
            measure: function (state) {
                var msr = 0;
                for (var irA = 0; irA < nregions; irA++) {
                    var regionA = data.regions[irA];
                    var cellA = state.region_to_grid[irA];
                    for (var irB = irA + 1; irB < nregions; irB++) {
                        var regionB = data.regions[irB];
                        var cellB = state.region_to_grid[irB];
                        msr += Math.pow(computeEuclideanDistance(regionA.center, regionB.center)
                                - computeEuclideanDistance(cellA.center, cellB.center), 2);
                    }
                }
                return msr;
            },
            sa_different: false,
            sa_measure: function (state) {
                return this.measure(state);
            }
        });
    }

    if (getAll || d3.select("#dist-neighbors-enabled").property("checked")) {
        var max;
        if (basicOnly) {
            max = 0;
        } else {
            var minD = computeEuclideanDistance(data.gridcells[0].center, data.gridcells[1].center);
            var maxD = computeEuclideanDistance(data.gridcells[0].center, data.gridcells[data.gridcells.length - 1].center);
            max = 0;
            neighborpairs.forEach(function (pair) {
                var regionA = pair[0];
                var regionB = pair[1];
                var rD = computeEuclideanDistance(regionA.center, regionB.center);
                max += Math.max(
                        Math.pow(rD - minD, 2),
                        Math.pow(rD - maxD, 2));

            });
        }
        settings.measures.push({
            name: "DIST N",
            interfacename: "Distance (nbr)",
            label: "dist-neighbors",
            weight: getAll ? 0 : Number(d3.select("#dist-neighbors-weight").property("value")),
            maxmeasure: max,
            invariance: {
                translation: true,
                rotation: true,
                scale: false,
                regions: false
            },
            measure: function (state) {
                var msr = 0;
                neighborpairs.forEach(function (pair) {
                    var regionA = pair[0];
                    var cellA = state.region_to_grid[regionA.index];
                    var regionB = pair[1];
                    var cellB = state.region_to_grid[regionB.index];
                    msr += Math.pow(computeEuclideanDistance(regionA.center, regionB.center)
                            - computeEuclideanDistance(cellA.center, cellB.center), 2);
                });
                return msr;
            },
            sa_different: false,
            sa_measure: function (state) {
                return this.measure(state);
            }
        });
    }

    if (getAll || d3.select("#dir-allpairs-enabled").property("checked")) {
        settings.measures.push({
            name: "DIR AP",
            interfacename: "Direction (all)",
            label: "dir-allpairs",
            weight: getAll ? 0 : Number(d3.select("#dir-allpairs-weight").property("value")),
            maxmeasure: getAll ? 0 : Math.PI * nregions * (nregions - 1) / 2,
            invariance: {
                translation: true,
                rotation: false,
                scale: true,
                regions: false
            },
            measure: function (state) {
                var msr = 0;
                for (var irA = 0; irA < nregions; irA++) {
                    var regionA = data.regions[irA];
                    var cellA = state.region_to_grid[irA];
                    for (var irB = irA + 1; irB < nregions; irB++) {
                        var regionB = data.regions[irB];
                        var cellB = state.region_to_grid[irB];
                        msr += Math.abs(computeSignedAngle(
                                regionA.center, regionB.center,
                                cellA.center, cellB.center));
                    }
                }
                return msr;
            },
            sa_different: false,
            sa_measure: function (state) {
                return this.measure(state);
            }
        });
    }

    if (getAll || d3.select("#dir-neighbors-enabled").property("checked")) {
        settings.measures.push({
            name: "DIR N",
            interfacename: "Direction (nbr)",
            label: "dir-neighbors",
            weight: getAll ? 0 : Number(d3.select("#dir-neighbors-weight").property("value")),
            maxmeasure: getAll ? 0 : Math.PI * neighborpairs.length,
            invariance: {
                translation: true,
                rotation: false,
                scale: true,
                regions: false
            },
            measure: function (state) {
                var msr = 0;
                neighborpairs.forEach(function (pair) {
                    var regionA = pair[0];
                    var cellA = state.region_to_grid[regionA.index];
                    var regionB = pair[1];
                    var cellB = state.region_to_grid[regionB.index];
                    msr += Math.abs(computeSignedAngle(
                            regionA.center, regionB.center,
                            cellA.center, cellB.center));
                });
                return msr;
            },
            sa_different: false,
            sa_measure: function (state) {
                return this.measure(state);
            }
        });
    }

    if (getAll || d3.select("#dir-orthoallpairs-enabled").property("checked")) {
        settings.measures.push({
            name: "DIR ORTH-AP",
            interfacename: "Compass Direction (all)",
            label: "dir-orthoallpairs",
            weight: getAll ? 0 : Number(d3.select("#dir-orthoallpairs-weight").property("value")),
            maxmeasure: getAll ? 0 : 2 * nregions * (nregions - 1) / 2,
            invariance: {
                translation: true,
                rotation: false,
                scale: true,
                regions: false
            },
            measure: function (state) {
                var msr = 0;
                for (var irA = 0; irA < nregions; irA++) {
                    var regionA = data.regions[irA];
                    var cellA = state.region_to_grid[irA];
                    for (var irB = irA + 1; irB < nregions; irB++) {
                        var regionB = data.regions[irB];
                        var cellB = state.region_to_grid[irB];
                        if ((regionA.center[0] - regionB.center[0]) * (cellA.center[0] - cellB.center[0]) < -0.00001) {
                            msr++;
                        }
                        if ((regionA.center[1] - regionB.center[1]) * (cellA.center[1] - cellB.center[1]) < -0.00001) {
                            msr++;
                        }
                    }
                }
                return msr;
            },
            sa_different: true,
            sa_measure: function (state) {
                var msr = 0;
                for (var irA = 0; irA < nregions; irA++) {
                    var regionA = data.regions[irA];
                    var cellA = state.region_to_grid[irA];
                    for (var irB = irA + 1; irB < nregions; irB++) {
                        var regionB = data.regions[irB];
                        var cellB = state.region_to_grid[irB];
                        if ((regionA.center[0] - regionB.center[0]) * (cellA.center[0] - cellB.center[0]) < -0.00001) {
                            msr += 1 + 0.1 * Math.abs(cellA.column - cellB.column);
                        }
                        if ((regionA.center[1] - regionB.center[1]) * (cellA.center[1] - cellB.center[1]) < -0.00001) {
                            msr += 1 + 0.1 * Math.abs(cellA.row - cellB.row);
                        }
                    }
                }
                return msr;
            }
        });
    }

    if (getAll || d3.select("#dir-orthoneighbors-enabled").property("checked")) {
        settings.measures.push({
            name: "DIR ORTH-N",
            interfacename: "Compass Direction (nbr)",
            label: "dir-orthoneighbors",
            weight: getAll ? 0 : Number(d3.select("#dir-orthoneighbors-weight").property("value")),
            maxmeasure: getAll ? 0 : 2 * neighborpairs.length,
            invariance: {
                translation: true,
                rotation: false,
                scale: true,
                regions: false
            },
            measure: function (state) {
                var msr = 0;
                neighborpairs.forEach(function (pair) {
                    var regionA = pair[0];
                    var cellA = state.region_to_grid[regionA.index];
                    var regionB = pair[1];
                    var cellB = state.region_to_grid[regionB.index];
                    if ((regionA.center[0] - regionB.center[0]) * (cellA.center[0] - cellB.center[0]) < -0.00001) {
                        msr++;
                    }
                    if ((regionA.center[1] - regionB.center[1]) * (cellA.center[1] - cellB.center[1]) < -0.00001) {
                        msr++;
                    }
                });
                return msr;
            },
            sa_different: true,
            sa_measure: function (state) {
                var msr = 0;
                neighborpairs.forEach(function (pair) {
                    var regionA = pair[0];
                    var cellA = state.region_to_grid[regionA.index];
                    var regionB = pair[1];
                    var cellB = state.region_to_grid[regionB.index];
                    if ((regionA.center[0] - regionB.center[0]) * (cellA.center[0] - cellB.center[0]) < -0.00001) {
                        msr += 1 + 0.1 * Math.abs(cellA.column - cellB.column);
                    }
                    if ((regionA.center[1] - regionB.center[1]) * (cellA.center[1] - cellB.center[1]) < -0.00001) {
                        msr += 1 + 0.1 * Math.abs(cellA.row - cellB.row);
                    }
                });
                return msr;
            }
        });
    }

    if (getAll || d3.select("#topo-broken-enabled").property("checked")) {
        settings.measures.push({
            name: "TOPO BREAKS",
            interfacename: "Split Neighbors",
            label: "topo-broken",
            weight: getAll ? 0 : Number(d3.select("#topo-broken-weight").property("value")),
            maxmeasure: getAll ? 0 : neighborpairs.length,
            invariance: {
                translation: true,
                rotation: true,
                scale: false,
                regions: false
            },
            measure: function (state) {
                var msr = 0;
                neighborpairs.forEach(function (pair) {
                    var regionA = pair[0];
                    var cellA = state.region_to_grid[regionA.index];
                    var regionB = pair[1];
                    var cellB = state.region_to_grid[regionB.index];
                    switch (data.gridmath.getCommonGeometry(cellA, cellB).length) {
                        case 0:
                            // broken
                            // NB: hopdistance should be > 1 now                            
                                msr += 1;
                            break;
                        case 1:
                            // point-adjacent
                            msr += 0.3;
                            break;
                        case 2:
                            // edge-adjacent
                            break;
                    }
                });
                return msr;
            },
            sa_different: true,
            sa_measure: function (state) {
                var msr = 0;
                neighborpairs.forEach(function (pair) {
                    var regionA = pair[0];
                    var cellA = state.region_to_grid[regionA.index];
                    var regionB = pair[1];
                    var cellB = state.region_to_grid[regionB.index];
                    switch (data.gridmath.getCommonGeometry(cellA, cellB).length) {
                        case 0:
                            // broken
                            // NB: hopdistance should be > 1 now
                            msr += 1 + 0.1 * (data.gridmath.hopDistance(cellA, cellB) - 1);
                            break;
                        case 1:
                            // point-adjacent
                            msr += 0.3;
                            break;
                        case 2:
                            // edge-adjacent
                            break;
                    }
                });
                return msr;
            }
        });
    }

    if (getAll || d3.select("#topo-false-enabled").property("checked")) {
        settings.measures.push({
            name: "TOPO FALSE",
            interfacename: "False neighbors",
            label: "topo-false",
            weight: getAll ? 0 : Number(d3.select("#topo-false-weight").property("value")),
            maxmeasure: getAll ? 0 : nregions * (nregions - 1) / 2 - neighborpairs.length,
            invariance: {
                translation: true,
                rotation: true,
                scale: false,
                regions: false
            },
            measure: function (state) {
                // check for false adjacencies
                function check(region, neighborcell, edgecontact) {
                    var neighborregion = state.grid_to_region[neighborcell.index];
                    if (neighborregion === null) {
                        return 0;
                    }
                    // check if regions are neighboring
                    var toponeighbors = false;
                    var tpindex = 0;
                    while (!toponeighbors && tpindex < region.topology.length) {
                        toponeighbors = region.topology[tpindex].neighbor === neighborregion;
                        tpindex++;
                    }
                    if (toponeighbors) {
                        // we're happy...
                        return 0;
                    } else {
                        // false adjacency
                        if (edgecontact) {
                            return 1;
                        } else {
                            return 0.3;
                        }
                    }
                }

                var msr = 0;
                // let's see how much we save on those we do find
                for (var ri = 0; ri < nregions; ri++) {
                    var region = data.regions[ri];
                    var cell = state.region_to_grid[ri];

                    var nbs = data.gridmath.getNeighbors(cell, true);
                    nbs.forEach(function (nb) {
                        msr += check(region, nb, data.gridmath.getCommonGeometry(cell, nb).length === 2);
                    });
                }
                return msr;
            },
            sa_different: false,
            sa_measure: function (state) {
                return this.measure(state);
            }
        });
    }

    if (getAll || d3.select("#shape-emd-enabled").property("checked")) {
        
        var nearest = [];
        if (!basicOnly) {
            for (var ci = 0; ci < ncells; ci++) {
                nearest.push(cells_by_distance(data.gridcells[ci]));
            }
        }
        var max = 0;
        if (!basicOnly) {
            for (var ci = 0; ci < ncells; ci++) {
                var cell = data.gridcells[ci];
                max += nearest[ci][nearest.length - 1].distance * cell.overlap;
                max += cell.spread;
            }
        }
        settings.measures.push({
            name: "SHAPE EMD",
            interfacename: "Shape",
            label: "shape-emd",
            weight: getAll ? 0 : Number(d3.select("#shape-emd-weight").property("value")),
            maxmeasure: getAll ? 0 : max,
            invariance: {
                translation: false,
                rotation: false,
                scale: false,
                regions: true
            },
            measure: function (state) {
                var msr = 0;
                for (var ci = 0; ci < ncells; ci++) {
                    var dAssigned = Number.NaN;
                    for (var i = 0; i < nearest[ci].length; i++) {
                        if (state.grid_to_region[nearest[ci][i].cell.index] !== null) {
                            dAssigned = nearest[ci][i].distance;
                            break;
                        }
                    }
                    msr += dAssigned * data.gridcells[ci].overlap;
                }
                for (var ri = 0; ri < nregions; ri++) {
                    if (state.region_to_grid[ri] !== null) {
                        msr += state.region_to_grid[ri].spread;
                    }
                }
                return msr;
            },
            sa_different: false,
            sa_measure: function (state) {
                return this.measure(state);
            }
        });
    }
 
    if (getAll || d3.select("#data-compact-enabled").property("checked")) {
        settings.measures.push({
            name: "DAT COMPACT",
            interfacename: "Compactness",
            label: "data-compact",
            weight: getAll ? 0 : Number(d3.select("#data-compact-weight").property("value")),
            maxmeasure: getAll ? 0 : nregions * (nregions - 1) * Math.sqrt(2) * 500,
            invariance: {
                translation: true,
                rotation: true,
                scale: false,
                regions: true
            },
            measure: function (state) {
                var msr = 0;
                for (var irA = 0; irA < nregions; irA++) {
                    var cellA = state.region_to_grid[irA];
                    for (var irB = irA + 1; irB < nregions; irB++) {
                        var cellB = state.region_to_grid[irB];
                        msr += computeEuclideanDistance(cellA.center, cellB.center);
                    }
                }
                return 2 * msr;
            },
            sa_different: false,
            sa_measure: function (state) {
                return this.measure(state);
            }
        });
    }

    if (getAll || d3.select("#data-alignhorz-enabled").property("checked")) {
        var max;
        if (basicOnly) {
            max = 0;
        } else {
            var ncols = data.grid.length;
            var nregionsleft = nregions;
            max = 0;
            while (ncols < nregionsleft) {
                max += ncols * (ncols - 1) / 2;
                nregionsleft -= ncols;
            }
            max += nregionsleft * (nregionsleft - 1) / 2;
        }
        settings.measures.push({
            name: "DAT ALIGN-H",
            interfacename: "Horizontal Alignment",
            label: "data-alignhorz",
            weight: getAll ? 0 : Number(d3.select("#data-alignhorz-weight").property("value")),
            maxmeasure: max,
            invariance: {
                translation: true,
                rotation: false,
                scale: true,
                regions: true
            },
            measure: function (state) {
                // NB: low values should be good, so flip it
                var pairsaligned = 0;
                // grid[col][row]
                for (var row = 0; row < data.grid[0].length; row++) {
                    var inrow = 0;
                    for (var col = 0; col < data.grid.length; col++) {
                        if (state.grid_to_region[data.grid[col][row].index] !== null) {
                            inrow++;
                        }
                    }
                    if (inrow > 1)
                        pairsaligned += inrow * (inrow - 1) / 2;
                }
                return this.maxmeasure - pairsaligned;
            },
            sa_different: false,
            sa_measure: function (state) {
                return this.measure(state);
            }
        });
    }

    if (getAll || d3.select("#data-alignvert-enabled").property("checked")) {
        var max;
        if (basicOnly) {
            max = 0;
        } else {
            var nrows = data.grid[0].length;
            var nregionsleft = nregions;
            max = 0;
            while (nrows < nregionsleft) {
                max += nrows * (nrows - 1) / 2;
                nregionsleft -= nrows;
            }
            max += nregionsleft * (nregionsleft - 1) / 2;
        }
        settings.measures.push({
            name: "DAT ALIGN-V",
            interfacename: "Vertical Alignment",
            label: "data-alignvert",
            weight: getAll ? 0 : Number(d3.select("#data-alignvert-weight").property("value")),
            maxmeasure: max,
            invariance: {
                translation: true,
                rotation: false,
                scale: true,
                regions: true
            },
            measure: function (state) {
                // NB: low values should be good, so flip it
                var pairsaligned = 0;
                // grid[col][row]
                for (var col = 0; col < data.grid.length; col++) {
                    var incol = 0;
                    for (var row = 0; row < data.grid[0].length; row++) {
                        if (state.grid_to_region[data.grid[col][row].index] !== null) {
                            incol++;
                        }
                    }
                    if (incol > 1)
                        pairsaligned += incol * (incol - 1) / 2;
                }
                return this.maxmeasure - pairsaligned;
            },
            sa_different: false,
            sa_measure: function (state) {
                return this.measure(state);
            }
        });
    }

    return settings;
}

function measureCurrentAssignment(settings, state, suppressclear) {

    if (!batch.active && !suppressclear) {
        debug.clear();
    }

    if (!hasValue(state)) {
        var allassigned = true;
        data.regions.forEach(function (r) {
            allassigned = allassigned && hasValue(r.cell);
        });

        if (false && !allassigned) {
            return;
        }

        state = readSAState();
    }

    if (!hasValue(settings)) {
        settings = getSettings();
    }

    var penalty = {};
    penalty.total = 0;
    penalty.sa_total = 0;
    var totweight = 0;
    settings.measures.forEach(function (measure) {
        var msr = measure.measure(state);
        var sa_msr = measure.sa_measure(state);
        penalty[measure.name] = {
            value: msr,
            normvalue: msr * 100 / measure.maxmeasure,
            //weightvalue: measure.weight * msr * 100 / measure.maxmeasure,
            sa_value: sa_msr,
            sa_normvalue: sa_msr * 100 / measure.maxmeasure//,
                    //sa_weightvalue: measure.weight * sa_msr * 100 / measure.maxmeasure
        };
        penalty.total += measure.weight * msr * 100 / measure.maxmeasure;
        penalty.sa_total += measure.weight * sa_msr * 100 / measure.maxmeasure;
        totweight += measure.weight;
    });
    penalty.total /= totweight;
    penalty.sa_total /= totweight;

    if (batch.active || batchLikeOutput) {
        var builder = batch.active ? batch.result_data : debug;

        if (!batch.active) {
            builder.append("map\twhitespace");
            settings.measures.forEach(function (measure) {
                builder.append("\t" + measure.name + " (value)");
                builder.append("\t" + measure.name + " (norm)");
                if (measure.sa_different) {
                    builder.append("\t" + measure.name + "/SA (value)");
                    builder.append("\t" + measure.name + "/SA (norm)");
                }
            });
            builder.appendLine("");
            builder.append("MAP\tWS");
        }

        settings.measures.forEach(function (measure) {
            var pen = penalty[measure.name];
            builder.append("\t" + pen.value + "\t" + pen.normvalue);
            if (measure.sa_different) {
                builder.append("\t" + pen.sa_value + "\t" + pen.sa_normvalue);
            }
        });
        builder.append("\t" + penalty.total);
        builder.append("\t" + penalty.sa_total);
    } else {
        debug.appendLine("TOTAL COST: \t" + penalty.total);
        settings.measures.forEach(function (measure) {
            var pen = penalty[measure.name];
            debug.appendLine("  " + measure.name + ": \t" + measure.weight
                    + "\t" + pen.normvalue
                    + "\t<-- \t" + pen.value + "\t * 100 / \t" + measure.maxmeasure);
            if (measure.sa_different) {
                debug.appendLine("  " + measure.name + "/SA: \t" + measure.weight
                        + "\t" + pen.sa_normvalue
                        + "\t<-- \t" + pen.sa_value + "\t * 100 / \t" + measure.maxmeasure);
            }
        });
    }
}