/*
    © Copyright 2016 Wouter Meulemans 
    
    This file is part of SMWGComputation.

    SMWGComputation is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SMWGComputation is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with SMWGComputation.  If not, see <http://www.gnu.org/licenses/>.
*/

/* global data, debug, batch, visual, lp */

function hillClimb() {
    recompute(false, true);
}

function recompute(suppressclear, hconly) {
    if (!hasValue(data.grid)) {
        return;
    }
    if (!hasValue(data.regions)) {
        return;
    }

    if (!suppressclear)
        debug.clear();

    if (data.regions.length > data.gridcells.length) {
        debug.appendLine("Not enough grid cells: " + data.gridcells.length + " provided, " + data.regions.length + " needed");
        return;
    }

    var settings = getSettings();

    computeAssignment(settings, hconly);

    outputAssignment();
    visual.visualizeAssignment();
}

function computeAssignment(settings, hconly) {
    var nregions = data.regions.length;
    var ncells = data.gridcells.length;

    function compute_state_penalty(state, settings) {
        var penalty = 0;
        var totweight = 0;
        settings.measures.forEach(function (measure) {
            if (measure.weight > 0.0001) {
                var samsr = measure.sa_measure(state);
                penalty += measure.weight * samsr * 100 / measure.maxmeasure;
                totweight += measure.weight;
            }
        });
        penalty /= totweight;
        return penalty;
    }

    function random_state() {
        var state = {
            grid_to_region: [],
            region_to_grid: [],
            penalty: null
        };

        var emptycells = [];
        for (var i = 0; i < ncells; i++) {
            emptycells.push(i);
            state.grid_to_region[i] = null;
        }
        for (var i = 0; i < nregions; i++) {
            var rc = Math.floor(Math.random() * emptycells.length);
            state.grid_to_region[emptycells[rc]] = data.regions[i];
            state.region_to_grid[i] = data.gridcells[emptycells[rc]];
            emptycells[rc] = emptycells[emptycells.length - 1];
            emptycells.pop();
        }

        state.penalty = compute_state_penalty(state, settings);
        return state;
    }

    function clone_state(state) {
        return {
            grid_to_region: state.grid_to_region.slice(),
            region_to_grid: state.region_to_grid.slice(),
            penalty: state.penalty
        };
    }

    function temperature(k, k_max) {
        return 1 - k / k_max;
    }

    var count_nonimproving = 0;
    function accept(penalty_old, penalty_new, temperature) {
        if (penalty_new < penalty_old) {
            return true;
        } else {
            var testval = Math.pow(Math.E, (penalty_old - penalty_new) / temperature);
            var acceptnew = testval >= Math.random();
            if (acceptnew) {
                count_nonimproving++;
            }
            return acceptnew;
        }
    }

    function get_neighbor(state, temperature) {
        var neighbor = clone_state(state);

        // pick a cell and swap it with one of its neighbors
        var region = data.regions[Math.floor(Math.random() * nregions)];
        var cell = neighbor.region_to_grid[region.index];

        // jump distance depends on temperature
        var d = 1 + Math.floor(Math.pow(temperature, 2) * Math.max(data.grid[0].length, data.grid.length));

        var rowmin = Math.max(cell.row - d, 0);
        var rowmax = Math.min(cell.row + d, data.grid[0].length - 1);
        var colmin = Math.max(cell.column - d, 0);
        var colmax = Math.min(cell.column + d, data.grid.length - 1);

        var swapcell;
        do {
            var random_row = rowmin + Math.floor((rowmax + 1 - rowmin) * Math.random());
            var random_col = colmin + Math.floor((colmax + 1 - colmin) * Math.random());
            swapcell = data.grid[random_col][random_row];
        } while (swapcell === cell);

        var swapregion = neighbor.grid_to_region[swapcell.index];

        neighbor.grid_to_region[cell.index] = swapregion;
        neighbor.grid_to_region[swapcell.index] = region;

        neighbor.region_to_grid[region.index] = swapcell;
        if (swapregion !== null) {
            neighbor.region_to_grid[swapregion.index] = cell;
        }

        neighbor.penalty = compute_state_penalty(neighbor, settings);

        return neighbor;
    }

    var current_state;

    var allassigned = true;
    data.regions.forEach(function (r) {
        allassigned = allassigned && hasValue(r.cell);
    });
    if (!allassigned) {
        current_state = null;
    } else {
        current_state = readSAState();
        current_state.penalty = compute_state_penalty(current_state, settings);
    }

    var init_starttime = performance.now();
    var state;
    var reusetime = false;
    var lp_init_done = false;
    
        switch (settings.sa.init) {
            case "current-random":
                if (current_state !== null) {
                    state = clone_state(current_state);
                    break;
                }
                // run into next case
            case "random":
                state = random_state();
                break;
            case "current-lp":
                if (current_state !== null) {
                    state = clone_state(current_state);
                    break;
                }
                // run into next case
            case "lp":
                lp_init_done = true;
                if (batch.active && batch.lp_init_state !== null) {
                    state = clone_state(batch.lp_init_state);
                    reusetime = true;
                } else {
                    state = random_state();
                    lp.compute(false, state, batch.active);
                    if (batch.active) {
                        batch.lp_init_state = clone_state(state);
                    }
                }
                state.penalty = compute_state_penalty(state, settings);
                break;
        }
    
    var init_endtime = performance.now();

    if (batch.active) {
        if (reusetime) {
            batch.result_data.append("\t" + batch.lp_init_time);
        } else {
            batch.result_data.append("\t" + (init_endtime - init_starttime));
            batch.lp_init_time = (init_endtime - init_starttime);
        }
    } else {
        debug.appendLine("INIT");
        debug.appendLine("in " + (init_endtime - init_starttime) + "ms");
    }

    var locindiv = settings.getMeasureByName("LOC INDIV");
    var locindiv_only_and_lp_init = false;
    if (lp_init_done && locindiv.weight > 0.0001) {
        locindiv_only_and_lp_init = true;
        settings.measures.forEach(function (measure) {
            locindiv_only_and_lp_init = locindiv_only_and_lp_init
                    && (measure === locindiv || measure.weight <= 0.0001);
        });
    }

    if (!hconly && !locindiv_only_and_lp_init) {
        var sa_starttime = performance.now();

        for (var round = 0; round < settings.sa.rounds; round++) {
            var T = temperature(round, settings.sa.rounds);
            var nbr = get_neighbor(state, T);
            if (accept(state.penalty, nbr.penalty, settings.sa.tempscale * T)) {
                state = nbr;
            }
        }

        var sa_endtime = performance.now();

        if (batch.active) {
            batch.result_data.append("\t" + (sa_endtime - sa_starttime));
        } else {
            debug.appendLine("SA RESULT (" + count_nonimproving + " nonimproving steps kept)");
            debug.appendLine("in " + (sa_endtime - sa_starttime) + "ms");
        }
    } else if (batch.active) {
        batch.result_data.append("\t");
    }

    if (settings.sa.hillclimbing && !locindiv_only_and_lp_init) {

        var hc_starttime = performance.now();

        var swapsmade = true;
        var hcrounds = 0;
        var timebreak = false;
        while (swapsmade) {
            if ((settings.sa.hillclimbingtime > 0 && (performance.now() - hc_starttime) >= settings.sa.hillclimbingtime * 1000)) {
                timebreak = true;
                break;
            }

            swapsmade = false;
            // try all possible swaps, and keep it if the penalty decreases
            for (var cellAindex = 0; cellAindex < ncells; cellAindex++) {
                if ((settings.sa.hillclimbingtime > 0 && (performance.now() - hc_starttime) >= settings.sa.hillclimbingtime * 1000)) {
                    timebreak = true;
                    break;
                }

                var rA = state.grid_to_region[cellAindex];
                for (var cellBindex = cellAindex + 1; cellBindex < ncells; cellBindex++) {
                    var rB = state.grid_to_region[cellBindex];
                    if (rA !== null || rB !== null) {
                        // swap
                        state.grid_to_region[cellAindex] = rB;
                        state.grid_to_region[cellBindex] = rA;

                        if (rA !== null)
                            state.region_to_grid[rA.index] = data.gridcells[cellBindex];
                        if (rB !== null)
                            state.region_to_grid[rB.index] = data.gridcells[cellAindex];

                        var newpenalty = compute_state_penalty(state, settings);
                        if (state.penalty > newpenalty) {
                            // keep
                            state.penalty = newpenalty;
                            rA = rB;
                            swapsmade = true;
                        } else {
                            // swap back
                            state.grid_to_region[cellAindex] = rA;
                            state.grid_to_region[cellBindex] = rB;
                            if (rA !== null)
                                state.region_to_grid[rA.index] = data.gridcells[cellAindex];
                            if (rB !== null)
                                state.region_to_grid[rB.index] = data.gridcells[cellBindex];
                        }
                    }
                }
            }

            hcrounds++;
        }

        var hc_endtime = performance.now();

        if (batch.active) {
            batch.result_data.append("\t" + (hc_endtime - hc_starttime) + "\t" + hcrounds + "\t" + (timebreak ? 1 : 0));
        } else {
            debug.appendLine("HC RESULT");
            debug.appendLine("in " + (hc_endtime - hc_starttime) + "ms, " + hcrounds + " rounds" + (timebreak ? " // out of time!" : ""));
        }
    } else if (batch.active) {
        batch.result_data.append("\t\t\t");
    }

    if (settings.sa.invaropt && !locindiv_only_and_lp_init) {

        var lp_start = performance.now();

        // check invariances
        var alltranslationinvariant = true;
        var allrotationinvariant = true;
        var allagnostic = true;

        settings.measures.forEach(function (measure) {
            if (measure.weight > 0.0001) {
                alltranslationinvariant = alltranslationinvariant
                        && measure.invariance.translation;
                allrotationinvariant = allrotationinvariant
                        && measure.invariance.rotation;
                allagnostic = allagnostic
                        && (measure === locindiv || measure.invariance.regions);
            }
        });

        var nrows = data.grid[0].length;
        var ncolumns = data.grid.length;

        var translates = [null];
        var rotates = [null];
        var flips = [null];
        if (alltranslationinvariant) {
            var minRow = Number.POSITIVE_INFINITY;
            var maxRow = Number.NEGATIVE_INFINITY;
            var minCol = Number.POSITIVE_INFINITY;
            var maxCol = Number.NEGATIVE_INFINITY;
            for (var ri = 0; ri < nregions; ri++) {
                var cell = state.region_to_grid[ri];
                minRow = Math.min(cell.row, minRow);
                maxRow = Math.max(cell.row, maxRow);
                minCol = Math.min(cell.column, minCol);
                maxCol = Math.max(cell.column, maxCol);
            }

            function add(dr, dc) {
                var DR = dr;
                var DC = dc;
                translates.push(function (cell) {
                    return data.getGridCell(cell.column + DC, cell.row + DR);
                });
            }

            for (var dr = 1; dr <= minRow; dr++) {
                add(-dr, 0);
                for (var dc = 1; dc <= minCol; dc++) {
                    add(-dr, -dc);
                }
                for (var dc = 1; dc <= ncolumns - 1 - maxCol; dc++) {
                    add(-dr, dc);
                }
            }
            for (var dr = 1; dr <= nrows - 1 - maxRow; dr++) {
                add(dr, 0);
                for (var dc = 1; dc <= minCol; dc++) {
                    add(dr, -dc);
                }
                for (var dc = 1; dc <= ncolumns - 1 - maxCol; dc++) {
                    add(dr, dc);
                }
            }
            for (var dc = 1; dc <= minCol; dc++) {
                add(0, -dc);
            }
            for (var dc = 1; dc <= ncolumns - 1 - maxCol; dc++) {
                add(0, dc);
            }
        }
        if (allrotationinvariant) {
            // find out which rotate/flip is optimal?
            rotates.push(function (cell) {
                // rotate 90
                return data.getGridCell(
                        Math.floor(ncolumns / 2) - (cell.row - Math.floor(nrows / 2)),
                        Math.floor(nrows / 2) + (cell.column - Math.floor(ncolumns / 2)));
            });
            rotates.push(function (cell) {
                // rotate 180
                return data.getGridCell(ncolumns - 1 - cell.column, nrows - 1 - cell.row);
            });
            rotates.push(function (cell) {
                // rotate 270
                return data.getGridCell(
                        Math.floor(ncolumns / 2) + (cell.row - Math.floor(nrows / 2)),
                        Math.floor(nrows / 2) - (cell.column - Math.floor(ncolumns / 2)));
            });

            flips.push(function (cell) {
                // horizontal flip
                return data.getGridCell(cell.column, nrows - 1 - cell.row);
            });
            // while rotations and flips are in the same category, only one flip suffices
//            flips.push(function (cell) {
//                // vertical flip
//                return data.getGridCell(ncolumns - 1 - cell.column, nrows);
//            });
        }

        if (rotates.length + flips.length + translates.length > 3) {
            // translate, flip, rotate
            var best_mod = state;
            var best_loc = locindiv.measure(state);
            function tryMod(translate, rotate, flip) {
                if (translate === null && rotate === null && flip === null) {
                    return;
                }
                var mod = clone_state(state);
                var valid = true;
                for (var ri = 0; valid && ri < nregions; ri++) {
                    var cell = state.region_to_grid[ri];
                    mod.grid_to_region[cell.index] = null;
                }
                for (var ri = 0; valid && ri < nregions; ri++) {
                    var cell = state.region_to_grid[ri];
                    if (translate !== null) {
                        cell = translate(cell);
                    }
                    if (cell !== null && flip !== null) {
                        cell = flip(cell);
                    }
                    if (cell !== null && rotate !== null) {
                        cell = rotate(cell);
                    }
                    if (cell !== null) {
                        mod.region_to_grid[ri] = cell;
                        mod.grid_to_region[cell.index] = data.regions[ri];
                    } else {
                        valid = false;
                    }
                }

                if (valid) {
                    var loc = locindiv.measure(mod);
                    if (loc < best_loc) {
                        best_mod = mod;
                        best_loc = loc;
                    }
                }
            }

            translates.forEach(function (T) {
                rotates.forEach(function (R) {
                    flips.forEach(function (F) {
                        tryMod(T, R, F);
                    });
                });
            });

            state = best_mod;
        }

        if (allagnostic) {
            lp.shuffle(state, true);
        }

        var lp_end = performance.now();

        if (batch.active) {
            batch.result_data.append("\t" + (lp_end - lp_start));
        } else {
            debug.appendLine("INVARIANCE OPT " + translates.length + " " + rotates.length + " " + flips.length);
            debug.appendLine("in " + (lp_end - lp_start) + "ms");
        }
    } else if (batch.active) {
        batch.result_data.append("\t");
    }

    if (settings.sa.maintainbest && current_state !== null && current_state.penalty < state.penalty) {
        //debug.appendLine("REVERTING");
        state = current_state;
    }

    if (!batch.active)
        debug.appendLine("");

    measureCurrentAssignment(settings, state, true);

    if (!batch.active) {
        for (var cellindex = 0; cellindex < ncells; cellindex++) {

            if (state.grid_to_region[cellindex] !== null) {
                data.gridcells[cellindex].region = state.grid_to_region[cellindex];
                state.grid_to_region[cellindex].cell = data.gridcells[cellindex];
            } else {
                data.gridcells[cellindex].region = undefined;
            }
        }
    }

    return state;
}

function readSAState() {
    var state = {
        grid_to_region: [],
        region_to_grid: [],
        penalty: null
    };

    for (var i = 0; i < data.gridcells.length; i++) {
        state.grid_to_region[i] = null;
    }
    for (var i = 0; i < data.regions.length; i++) {
        var rg = data.regions[i];
        if (hasValue(rg.cell)) {
            state.grid_to_region[rg.cell.index] = rg;
            state.region_to_grid[i] = rg.cell;
        } else {
            state.region_to_grid[i] = null;
        }
    }
    return state;
}