/*
    © Copyright 2016 Wouter Meulemans 
    
    This file is part of SMWGComputation.

    SMWGComputation is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SMWGComputation is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with SMWGComputation.  If not, see <http://www.gnu.org/licenses/>.
*/

/* global d3, data */

function hasValue(x) {
    return x !== undefined && x !== null;
}

function computeEuclideanDistance(p, q) {
    var dx = p[0] - q[0];
    var dy = p[1] - q[1];
    return Math.sqrt(dx * dx + dy * dy);
}

function computeSquaredEuclideanDistance(p, q) {
    var dx = p[0] - q[0];
    var dy = p[1] - q[1];
    return dx * dx + dy * dy;
}

// either p1->p2 direction vs q1->q2 direction
// or with two parameters, just the vectors p1, p2
function computeAngle(p1, p2, q1, q2) {
    var angle = computeSignedAngle(p1, p2, q1, q2);
    if (angle < 0) {
        angle = Math.PI * 2 + angle;
    }
    return angle;
}

function computeSignedAngle(p1, p2, q1, q2) {
    var xp, yp;
    var xq, yq;
    if (q1 === undefined) {
        // two param        
        xp = p1[0];
        yp = p1[1];

        xq = p2[0];
        yq = p2[0];
    } else {
        // four param        
        xp = p2[0] - p1[0];
        yp = p2[1] - p1[1];

        xq = q2[0] - q1[0];
        yq = q2[1] - q1[1];
    }

    var lenp = Math.sqrt(Math.pow(xp, 2) + Math.pow(yp, 2));
    if (lenp > 0.00001) {
        xp /= lenp;
        yp /= lenp;
    }

    var lenq = Math.sqrt(Math.pow(xq, 2) + Math.pow(yq, 2));
    if (lenq > 0.00001) {
        xq /= lenq;
        yq /= lenq;
    }

    var sign = 1;
    // crossproduct
    if (xp * yq - yp * xq < 0) {
        sign = -1;
    }

    var dotp = xp * xq + yp * yq;
    if (dotp <= -1) {
        return sign * Math.PI;
    } else if (dotp >= 1) {
        return 0;
    } else {
        return sign * Math.acos(dotp);
    }
}

function computeArea(polygon) {
    var area = 0;
    for (var i = 1; i < polygon.length; i++) {
        area += polygon[i - 1][0] * polygon[i][1]
                - polygon[i - 1][1] * polygon[i][0];
    }
    area += polygon[polygon.length - 1][0] * polygon[0][1]
            - polygon[polygon.length - 1][1] * polygon[0][0];
    return Math.abs(area / 2);
}

function computeMidpoint(polygon) {
    var A = 0;
    var Cx = 0;
    var Cy = 0;

    var p = polygon[polygon.length - 1];
    var q, i;
    for (i = 0; i < polygon.length; i++) {
        q = polygon[i];
        var cross = p[0] * q[1] - q[0] * p[1];
        A += cross;
        Cx += (p[0] + q[0]) * cross;
        Cy += (p[1] + q[1]) * cross;
        p = q;
    }
    A *= 3;
    Cx = Cx / A;
    Cy = Cy / A;
    return [Cx, Cy];
}

function getEmptyBox() {
    return {
        left: Number.POSITIVE_INFINITY,
        right: Number.NEGATIVE_INFINITY,
        top: Number.POSITIVE_INFINITY,
        bottom: Number.NEGATIVE_INFINITY,
        width: function () {
            return this.right - this.left;
        },
        height: function () {
            return this.bottom - this.top;
        },
        contains: function (p) {
            return this.left <= p[0] && p[0] <= this.right
                    && this.top <= p[1] && p[1] <= this.bottom;
        },
        buffer: function (b) {
            this.left -= b;
            this.right += b;
            this.top -= b;
            this.bottom += b;
        }
    };
}

function getBoundingBoxOfRegions(rgs, frame) {
    var box = getEmptyBox();
    rgs.forEach(function (r) {
        r.parts.forEach(function (p) {
            p.geometry.forEach(function (v) {
                if (!hasValue(frame) || frame.contains(v)) {
                    box.left = Math.min(box.left, v[0]);
                    box.right = Math.max(box.right, v[0]);
                    box.top = Math.min(box.top, v[1]);
                    box.bottom = Math.max(box.bottom, v[1]);
                }
            });
        });
    });
    return box;
}

function getBoundingBoxOfParts(prts, frame) {
    var box = getEmptyBox();
    prts.forEach(function (p) {
        p.geometry.forEach(function (v) {
            if (!hasValue(frame) || frame.contains(v)) {
                box.left = Math.min(box.left, v[0]);
                box.right = Math.max(box.right, v[0]);
                box.top = Math.min(box.top, v[1]);
                box.bottom = Math.max(box.bottom, v[1]);
            }
        });
    });
    return box;
}

function getBoundingBoxOfGeometry(geom, frame) {
    var box = getEmptyBox();
    geom.forEach(function (v) {
        if (!hasValue(frame) || frame.contains(v)) {
            box.left = Math.min(box.left, v[0]);
            box.right = Math.max(box.right, v[0]);
            box.top = Math.min(box.top, v[1]);
            box.bottom = Math.max(box.bottom, v[1]);
        }
    });
    return box;
}

function cells_by_distance(cell) {
    var result = [];
    for (var ci = 0; ci < data.gridcells.length; ci++) {
        var cell2 = data.gridcells[ci];
        result.push({
            cell: cell2,
            distance: computeEuclideanDistance(cell.center, cell2.center)
        });
    }
    result.sort(function (a, b) {
        return a.distance - b.distance;
    });
    return result;
}

function summary(vals) {
    vals.sort(function (a, b) {
        return a - b;
    });
    var tot = 0;
    vals.forEach(function (v) {
        tot += v;
    });
    var avg = tot / vals.length;
    var sumofsquares = 0;
    vals.forEach(function (v) {
        sumofsquares += Math.pow(v - avg, 2);
    });
    return {
        min: vals[0],
        max: vals[vals.length - 1],
        avg: avg,
        stddev: Math.sqrt(sumofsquares / vals.length),
        median: vals[Math.ceil(vals.length / 2)],
        pct75: vals[Math.ceil(3 * vals.length / 4)],
        pct25: vals[Math.ceil(1 * vals.length / 4)]
    };
}

function selectAllTextIn(containerid) {
    var node = document.getElementById(containerid);
    if (document.selection) {
        var range = document.body.createTextRange();
        range.moveToElementText(node);
        range.select();
    } else if (window.getSelection) {
        var range = document.createRange();
        range.selectNodeContents(node);
        window.getSelection().removeAllRanges();
        window.getSelection().addRange(range);
    }
}